"""
apkg packaging file cache
"""

import json
from pathlib import Path
from typing import Dict, Iterable, List, Mapping, Union

from apkg import __version__
from apkg.log import getLogger
from apkg.util.common import CacheableEntry, hash_file, hash_path


log = getLogger(__name__)


CacheEntry = Union[str, int, bool, List['CacheEntry'], Dict[str, 'CacheEntry']]


class FileMissingError(Exception):
    pass


class ChecksumError(Exception):
    pass


def file_checksum(*paths) -> str:
    return hash_file(*paths).hexdigest()[:20]


def path_checksum(*paths) -> str:
    return hash_path(*paths).hexdigest()[:20]


def enabled_str(enabled: bool) -> str:
    return 'ENABLED' if enabled else 'DISABLED'


class ProjectCache:
    def __init__(self, project):
        self.project = project
        self.loaded = False
        self.cache = {}
        self.checksum = None

    def save(self) -> None:
        self.cache['__version__'] = __version__
        json.dump(self.cache, self.project.path.cache.open('w'))

    def load(self) -> None:
        cache_path = self.project.path.cache
        if not cache_path.exists():
            log.verbose("cache not found: %s", cache_path)
            return
        log.verbose("loading cache: %s", cache_path)

        cache = json.load(cache_path.open('r'))
        version = cache.get('__version__', None)
        if version != __version__:
            log.warning("ignoring cache from different apkg version: "
                        "%s != %s", version, __version__)
            return
        self.cache = cache

    def _ensure_load(self) -> None:
        """
        ensure cache is loaded on demand and only once

        you don't need to call this directly
        """
        if self.loaded:
            return
        self.load()
        self.loaded = True

    def update(self, key: str, entry: CacheableEntry) -> None:
        """
        update cache entry
        """
        log.verbose("cache update: %s -> %s", key, entry)
        assert key
        self._ensure_load()
        self.cache[key] = cache_encode(entry)
        self.save()

    def get(self, key: str) -> CacheableEntry:
        """
        get and validate cache entry or None
        """
        log.verbose("cache query: %s", key)

        assert key
        self._ensure_load()
        entry = self.cache.get(key)
        if not entry:
            return None

        try:
            entry = cache_decode(entry)
        except FileMissingError as e:
            path = e.args[0]
            log.info("invalidating cache for %s due to missing file: %s",
                     key, path)
            self.delete(key)
            return None
        except ChecksumError as e:
            path = e.args[0]
            log.info("invalidating cache for %s due to failed checksum on: %s",
                     key, path)
            self.delete(key)
            path.unlink()
            return None
        return entry

    def delete(self, key: str) -> None:
        """
        delete cache entry
        """
        self.cache.pop(key, None)
        self.save()

    def enabled(self, *targets, cmd: str = None,
                use_cache: bool = True) -> bool:
        """
        helper to tell and log if caching should be enabled

        targets is a list of cache targets that must be enabled:

        * local: cache local files
        * remote: cache remote files
        * source: cache project source (requires VCS)

        optional use_cache utility argument to disable all cache
        """

        def enabled_result(value: bool) -> bool:
            en_str = enabled_str(value)
            if cmd:
                log.verbose("cache %s for %s", en_str, cmd)
            else:
                log.verbose("cache %s", en_str)
            return value

        if not use_cache:
            # all cache disabled
            return enabled_result(use_cache)

        r = True
        for target in targets:
            option = 'cache.%s' % target
            value = self.project.config_get(option)
            cache = True

            if value is not None:
                # set in project config
                if value and target == 'source' and not self.project.vcs:
                    # source cache requires VCS
                    msg = ("cache.{target} {en} in project config, "
                           "but VCS isn't available - cache.{target} {di}.\n"
                           "Please update your project config.").format(
                        target=target,
                        en=enabled_str(value),
                        di=enabled_str(False))
                    log.warning(msg)
                    cache = False
                else:
                    log.verbose("cache.%s %s in project config",
                                target, enabled_str(value))
                    cache = value

            else:
                # auto cache settings
                if target == 'source':
                    # source cache requires VCS
                    cache = bool(self.project.vcs)
                    log.verbose("cache.%s %s by default (VCS=%s)",
                                target, enabled_str(cache), self.project.vcs)
                else:
                    # other cache types are ENABLED by default
                    log.verbose("cache.%s %s by default",
                                target, enabled_str(True))

            # always go through all targets to get complete logs
            r = r and cache

        return enabled_result(r)


def cache_encode(entry: CacheableEntry) -> CacheEntry:
    """
    Encode entry for use in cache, suported formats are only:
        - int, bool, str
        - Path (stored with checksum)
        - list if all items are a supported format
        - dict where keys are str, values any supported format
    """
    if isinstance(entry, Mapping):
        result = {}
        for k, v in entry.items():
            if k.startswith('__'):
                raise ValueError(f"Underscore keys are not supported: {k}")
            result[k] = cache_encode(v)
        return result
    elif isinstance(entry, (int, str, bool)):  # str is Iterable
        return entry
    elif isinstance(entry, Iterable):
        return [cache_encode(v) for v in entry]
    elif isinstance(entry, Path):
        return {"path": str(entry), "__hash__": file_checksum(entry)}
    raise NotImplementedError


def cache_decode(entry: CacheEntry) -> CacheableEntry:
    """
    Decode and validate entry from cache, validating existence and checksum of
    path entries along the way.
    """
    if isinstance(entry, dict):
        if '__hash__' in entry:
            path = Path(entry['path'])
            checksum = entry['__hash__']
            if not path.exists():
                raise FileMissingError(path)
            if checksum != file_checksum(path):
                raise ChecksumError(path)
            return path
        return {k: cache_decode(v) for k, v in entry.items()}
    if isinstance(entry, list):
        return [cache_decode(v) for v in entry]
    return entry



import click

from apkg import adistro
from apkg import ex
from apkg.log import getLogger
from apkg import pkgstyle


log = getLogger(__name__)


@click.command(name='system-setup')
@click.option('-c', '--core', is_flag=True,
              help="install core packages for direct package builds  [default]")
@click.option('-I', '--isolated', is_flag=True,
              help="install packages for isolated package builds")
@click.option('-L', '--lint', is_flag=True,
              help="install packages for linting (apkg lint)")
@click.option('-a', '--all', is_flag=True,
              help="install all of above (-cIL)")
@click.option('-d', '--distro',
              help="override target distro  [default: current]")
@click.option('--ask/--no-ask', 'interactive',
              default=False, show_default=True,
              help="enable/disable interactive mode")
@click.help_option('-h', '--help',
                   help="show this help message")
def cli_system_setup(*args, **kwargs):
    """
    setup system for packaging

    Install native distro packages required for packaging.

    Select desired packages with a combination of

    -c / --core: core packages for direct package builds  [default]

    -I / --isolated`: packages for isolated package builds

    -L / --lint`: packages for linting (apkg lint)

    or

    -a / --all to select all above.

    Defaults to --core when no options are supplied.
    """
    return system_setup(*args, **kwargs)


def system_setup(
        core=False,
        isolated=False,
        lint=False,
        all=False,
        distro=None,
        interactive=False):
    """
    setup system for packaging
    """
    cats_opts = {
        'core': core,
        'isolated': isolated,
        'lint': lint,
    }
    if all:
        # install all categories
        cats = cats_opts.keys()
    else:
        # nstall only selected categories
        cats = [cat for cat, e in cats_opts.items() if e]
        if not cats:
            # default to --core
            cats = ['core']

    cats_txt = ", ".join(cats)
    log.bold("system setup: %s", cats_txt)

    distro = adistro.distro_arg(distro)
    log.info("target distro: %s", distro)

    style = pkgstyle.get_pkgstyle_for_distro(distro)
    if not style:
        raise ex.DistroNotSupported(distro=distro)
    log.info("target pkgstyle: %s", style.name)

    distro_reqs = getattr(style, 'DISTRO_REQUIRES', {})
    reqs = []
    for cat in cats:
        reqs += distro_reqs.get(cat, [])

    if reqs:
        pkgstyle.call_pkgstyle_fun(
            style, 'install_distro_packages',
            reqs,
            distro=distro,
            interactive=interactive)
    else:
        log.info("no distro packages required")

    log.success("system setup successful: %s", cats_txt)


APKG_CLI_COMMANDS = [cli_system_setup]



import click

from apkg import adistro
from apkg.commands.build import build
from apkg.commands.srcpkg import srcpkg as make_srcpkg
from apkg.commands.system_setup import system_setup
from apkg import ex
from apkg.log import getLogger
from apkg import pkgstyle
from apkg.util import common


log = getLogger(__name__)


@click.command(name="lint")
@click.argument('inputs', nargs=-1)
@click.option('-p', '--pedantic', is_flag=True,
              help="enable extra / all warnings and infos")
@click.option('-i', '--info', is_flag=True,
              help="give detailed info about warnings")
@click.option('-s', '--strict', is_flag=True,
              help="treat all messages as errors")
@click.option('-d', '--distro',
              help="override target distro  [default: current]")
@click.option('-L', '--lint-dep', is_flag=True,
              help="install linting dependencies on host"
                   " (apkg system-setup --lint)")
@click.option('--cache/--no-cache', default=True, show_default=True,
              help="enable/disable cache")
@click.option('-F', '--in-file', 'in_files', multiple=True,
              help="specify input file, '-' to read from stdin")
@click.help_option('-h', '--help',
                   help="show this help message")
def cli_lint(*args, **kwargs):
    """
    run native distro linter on packages

    Default: build packages from source and lint them

    You can supply files to lint as arguments or use --in-file.

    Use --pedantic to run extra / all checks.

    Use --info to get details explanations about individual messages.

    Use --strict to treat all messages as errors (useful for pedantic CI).

    To install linter packages on host system use --lint-dep which calls

        apkg system-setup --lint

    before linting.
    """
    r = lint(*args, **kwargs)
    if r != 0:
        raise ex.LintingFailed(
            fail=f"linter returned error code {r}")
    else:
        log.success("linting successful")
    return r


def lint(
        inputs=None,
        in_files=None,
        pedantic=False,
        info=False,
        strict=False,
        distro=None,
        lint_dep=False,
        cache=True):
    """
    run native distro linter on packages
    """
    log.bold("linting packages")

    distro = adistro.distro_arg(distro)
    log.info("target distro: %s", distro)

    ps = pkgstyle.get_pkgstyle_for_distro(distro)
    if not ps:
        raise ex.DistroNotSupported(distro=distro)
    log.info("target pkgstyle: %s", ps.name)

    if lint_dep:
        system_setup(lint=True)

    inputs = common.parse_inputs(inputs, in_files)

    if not inputs:
        # default: use srcpkg and build to get packages to lint
        inputs = make_srcpkg(
            distro=distro,
            cache=cache)
        inputs += build(
            distro=distro,
            cache=cache)

    try:
        result = pkgstyle.call_pkgstyle_fun(
            ps, 'lint',
            inputs,
            pedantic=pedantic,
            info=info,
            strict=strict,
            distro=distro)
    except ex.CommandNotFound as e:
        msg = str(e) + ("\n\nInstall lint deps:\n\n"
                        "    apkg system-setup --lint\n\n"
                        "or:\n\n"
                        "    apkg lint --lint-dep")
        raise ex.CommandNotFound(msg=msg)

    return result


APKG_CLI_COMMANDS = [cli_lint]



"""
Backward compatibility wrapper for debseries -> deb_series rename.

Please use deb_series module instead, this is going to be removed for 1.0.
"""  # noqa

from apkg.templatevars import deb_series
from apkg.log import getLogger


log = getLogger(__name__)


# called by apkg to get custom variables
def get_variables(env):
    log.warning("apkg.templatevars.debseries renamed to deb_series. "
                "Please use apkg.templatevars.deb_series instead.")
    return deb_series.get_variables(env)



from pathlib import Path
import shutil

import click
import yaml

from apkg import ex
from apkg.cli import cli
from apkg.util import common
from apkg.log import getLogger
from apkg.project import Project
from apkg.util.archive import get_archive_version
from apkg.util.run import run


log = getLogger(__name__)


@cli.command(name='make-archive', aliases=['ar'])
@click.option('-O', '--result-dir',
              help="put results into specified dir")
@click.option('-o', '--out-format', default='yaml', show_default=True,
              type=click.Choice(['yaml', 'list']),
              help="set output format")
@click.option('--cache/--no-cache', default=True, show_default=True,
              help="enable/distable cache")
@click.help_option('-h', '--help', help='show this help')
def cli_make_archive(*args, **kwargs):
    """
    create dev archive from current project state
    """
    out_format = kwargs.pop('out_format')

    results = make_archive(*args, **kwargs)

    if out_format == 'list':
        rlist = archive_dict2list(results)
        common.print_results(rlist)
    else:
        # YAML
        common.print_results_dict(results)

    return results


def make_archive(
        result_dir=None,
        cache=True,
        project=None):
    """
    create dev archive from current project state

    Use script specified by project.make_archive_script config option.
    """
    log.bold("creating dev archive")
    proj = project or Project()

    use_cache = proj.cache.enabled(
        'source', cmd='make_archive', use_cache=cache)
    if use_cache:
        cache_key = 'archive/dev/%s' % proj.checksum
        cached = common.get_cached_paths(proj, cache_key, result_dir)
        if cached:
            log.success("reuse cached archive: %s", cached['archive'])
            return cached

    script = proj.config_get('project.make_archive_script')
    if not script:
        msg = ("make-archive requires project.make_archive_script option to\n"
               "be set in project config to a script that creates project\n"
               "archive and prints its path to stdout in YAML like this:\n\n"
               "archive: pkg/archive/dev/foo-1.2.3.tar.gz\n\n"
               "Please update project config with required information:\n\n"
               "%s" % proj.path.config)
        raise ex.MissingRequiredConfigOption(msg=msg)

    log.info("running make_archive_script: %s", script)
    cmd_out = run(script, quiet=True)
    out = str(cmd_out.stdout)

    if result_dir:
        ar_base_path = Path(result_dir)
    else:
        ar_base_path = proj.path.dev_archive

    results = {}
    if proj.compat_level >= 6:
        try:
            results = yaml.safe_load(out)
        except Exception as e:
            msg = ("Failed to parse make_archive_script YAML output:\n\n"
                   "%s\n\nError:\n\n%s" % (out, e))
            raise ex.UnexpectedCommandOutput(msg=msg)
        if not isinstance(results, dict):
            msg = ("Invalid make_archive_script YAML output format:\n\n"
                   "%s\n\nExpected format example:\n\n"
                   "archive: pkg/archives/dev/foo-1.2.3.tar.gz" % (out))
            raise ex.UnexpectedCommandOutput(msg=msg)
    else:
        # use old undocumented behaviour when the last line of output was used
        msg = ("using legacy make_archive_script format (last line)"
               " due to compat level %s" % proj.compat_level)
        log.info(msg)
        lines = out.split('\n')
        archive_path = Path(lines[-1])
        results["archive"] = archive_path

    sanitize_archive_output(results)

    archive_path = results.get('archive')
    if not archive_path:
        msg="make_archive_script didn't return archive:\n\n%s" % out
        raise ex.UnexpectedCommandOutput(msg=msg)

    # copy all to result dir (updates results in-place)
    copy_archives(results, ar_base_path)

    archive_path = results.get('archive')
    log.success("made archive: %s", archive_path)

    if use_cache:
        proj.cache.update(cache_key, results)
    return results


def sanitize_archive_output(output: dict):
    # convert paths to pathlib.Path
    archive = output.get('archive')
    if archive:
        if not isinstance(archive, Path):
            archive = Path(archive)
            output['archive'] = archive
        if archive and not archive.exists():
            raise ex.ArchiveNotFound(ar=archive)
    components = output.get('components')
    if components:
        for comp_name, comp_path in list(components.items()):
            if not isinstance(comp_path, Path):
                comp_path = Path(comp_path)
                components[comp_name] = comp_path
            if not comp_path.exists():
                raise ex.ArchiveNotFound(
                    msg="Archive component '%s' not found: %s" %
                    (comp_name, comp_path))
    # provide explicit version when possible
    if 'version' not in output and archive:
        output['version'] = get_archive_version(archive)
    # version is always str (can occasionally get int or float)
    version = output.get('version')
    if version and not isinstance(version, str):
        output['version'] = str(version)


def archive_dict2list(output: dict) -> list:
    results = []
    archive = output.get('archive')
    if archive:
        results.append(archive)
    components = output.get('components')
    if components:
        results += components.values()
    return results


def copy_archives(output: dict, destdir: Path):
    """
    copy archive and all components to destdir
    """
    archive = output.get('archive')
    if archive:
        output['archive'] = copy_archive(archive, destdir)
    components = output.get('components')
    if components:
        for comp_name, comp_path in list(components.items()):
            components[comp_name] = copy_archive(
                comp_path, destdir, txt="'%s' component" % comp_name)


def copy_archive(source, destdir, name=None, txt='archive'):
    if name is None:
        name = source.name

    dest = destdir / name

    if source != dest:
        log.info("copying %s to: %s", txt, dest)
        destdir.mkdir(parents=True, exist_ok=True)
        shutil.copy(source, dest)

    return dest


APKG_CLI_COMMANDS = [cli_make_archive]



import re

from apkg import ex


# subgroups:
# 1) name
# 2) name-version separator ('-' or '_')
# 3) version (including release)
RE_NVR = r'^(.+?)([-_])(v?\d+(?:\.\d+)*(?:.+?)?)$'


def split_archive_ext(archive_fn):
    """
    split archive file into base name and extension such as '.tar.gz'

    return (name, extension)
    """
    name, _, ext = archive_fn.rpartition('.')
    if not name:
        return archive_fn, ''
    ext = '.%s' % ext
    if name.endswith('.tar'):
        name, _, _ = name.rpartition('.')
        ext = '.tar%s' % ext

    return name, ext


def split_archive_fn(archive_fn):
    """
    split archive file name into individual parts

    return (name, separator, version, extension)
    """
    name, ext = split_archive_ext(archive_fn)
    r = re.match(RE_NVR, name)
    if r:
        return r.groups() + (ext,)

    msg = "unable to parse version from archive file name: %s" % archive_fn
    raise ex.ParsingFailed(msg=msg)


def parse_version(version_str):
    """
    parse version from common version strings

    currently only strips v from vX.Y.Z
    """
    if version_str.startswith('v'):
        return version_str[1:]
    return version_str



"""
apkg helpers for distro parsing and matching

standard distro module is used for getting current distro
"""
from abc import ABC, abstractmethod
import re

try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property
from packaging.specifiers import SpecifierSet
from packaging import version

import distro

from apkg.log import getLogger
from apkg.util import common


log = getLogger(__name__)


RE_VERSION_RULE = r'(\w+)\s*((?:==|!=|~=|<=?|>=?)\s*\d+.*)'


class Distro:
    """
    represent distro by id and (optional) version
    """
    def __init__(self, distro_str, aliases=None):
        self.orig = distro_str
        # idver format, i.e.: debian-10, ubuntu-21.04
        self.idver = distro2idver(self.orig)
        self.id, self.version = parse_idver(self.idver)
        self.names = [self.id]
        self.aliases = aliases or {}

    def match(self, *rules):
        """
        return True if self matches supplied distro rule(s)
        """
        drules = distro_rules(rules, aliases=self.aliases)
        return drules.match(self)

    def print(self, sep='-', short=False):
        s = self.id
        if short:
            s = s[:3]
        if not self.version:
            return s
        if sep:
            s += sep
        s += str(self.version)
        return s

    @cached_property
    def human(self):
        """
        return distro in human-friendly format, i.e.: 'fedora 10'
        """
        return self.print(sep=' ')

    @cached_property
    def tiny(self):
        """
        return distro in tiny format, i.e.: deb10, ubu21.04
        """
        return self.print(sep='', short=True)

    @cached_property
    def parsed_version(self):
        return self.version and version.parse(self.version)

    def __str__(self):
        return self.human

    def __repr__(self):
        return "<Distro('%s')>" % self.idver


class DistroRuleBase(ABC):
    """
    base class for distro rules
    """
    @abstractmethod
    def match(self, distro_):
        pass


class DistroRule(DistroRuleBase):
    """
    rule matching specific distro id and version range
    """
    def __init__(self, rule_str):
        self.rule_str = rule_str
        m = re.match(RE_VERSION_RULE, rule_str)
        if m:
            self.id = m.group(1)
            spec_str = m.group(2)
            self.version_spec = SpecifierSet(spec_str)
        else:
            self.id = rule_str
            self.version_spec = None

    def match(self, distro_):
        if self.id not in distro_.names:
            return False
        if not self.version_spec:
            return True
        if not distro_.version:
            # Consider an empty/missing version higher than anything else
            return all(specifier.operator in ('!=', '>', '>=')
                       for specifier in self.version_spec)
        return distro_.parsed_version in self.version_spec

    def __str__(self):
        return self.rule_str

    def __repr__(self):
        return "<DistroRule('%s')>" % self.rule_str


class DistroRuleAll(DistroRuleBase):
    """
    rule matching any distro
    """
    def match(self, _):
        return True

    def __str__(self):
        return 'all distros'

    def __repr__(self):
        return "<DistroRuleAll>"


class DistroRules(DistroRuleBase):
    """
    list of multiple distro rules
    """
    def __init__(self, rules):
        self.rules = rules

    def match(self, distro_):
        """
        return True when any single rule matches (OR)
        """
        for rule in self.rules:
            if rule.match(distro_):
                return True
        return False

    def __str__(self):
        return ' | '. join([r.rule_str for r in self.rules])

    def __repr__(self):
        return "<ApkgDistroRules(%s)>" % self.__str__()


def distro_rules(rules, aliases=None):
    """
    parse (a list of) distro rule string(s) into ApkgDistroRule(s)
    using distro aliases if supplied

    Params:
        rules: a distro rule string or a list of thereof
        aliases: distro aliases dict (optional)

    Return:
        * ApkgDistroRule for a string
        * ApkgDistroRules for a list of strings
    """
    if isinstance(rules, str):
        # single rule string
        rules = [rules]

    aliases = aliases or {}
    drules = []
    for rule in rules:
        if rule in aliases:
            r = aliases[rule]
        else:
            r = DistroRule(rule)
        drules.append(r)

    if len(drules) == 1:
        # single DistroRule
        return drules[0]
    else:
        # multiple DistroRules
        return DistroRules(drules)


def name2rule(name):
    """
    return distro rule based on distro-specific (file) name
    """
    rule_str, _, ver = name.partition('-')
    if ver:
        rule_str += ' == %s' % ver
    return DistroRule(rule_str)


def sort_by_name(objs, name_attr='name'):
    """
    sort objects by distro-specific (file) name

    Params:
        objs: list of objects with name_attr attribute
        name_attr: object attribute (distro name) to sort by

    Return:
        objs list ordered by name_attr which is assumed to be
        distro name such as 'debian' or 'ubuntu-20.04':

        * names with release first
        * primary sort by name
        * secondary sort by release desc (for determinism)
    """

    def sort_key(o):
        distro_, _, rls = getattr(o, name_attr).rpartition('-')
        try:
            rlsv = version.parse(rls)
        except version.InvalidVersion:
            rlsv = rls
        return distro_, common.SortReversor(rlsv)

    plain_objs = []
    rls_objs = []
    for o in objs:
        if '-' in getattr(o, name_attr):
            rls_objs.append(o)
        else:
            plain_objs.append(o)

    rls_objs.sort(key=sort_key)
    plain_objs.sort(key=lambda x: getattr(x, name_attr))

    return rls_objs + plain_objs


def parse_distro_aliases(config):
    """
    parse distro aliases from config fromat into a dict
    """
    aliases = {}
    for alias in config:
        name = alias.get('name')
        if not name:
            log.warning("Ignoring invalid distro alias:"
                        " missing required option: name")
            continue
        distro_ = alias.get('distro')
        if not name:
            log.warning("Ignoring invalid distro alias:"
                        " missing required option: distro")
            continue
        try:
            rules = distro_rules(distro_)
        except Exception as ex:
            log.warning("Ignoring invalid distro alias '%s':"
                        " invalid distro rule: %s",
                        name, str(ex))
            continue
        aliases[name] = rules
    return aliases


def distro_arg(distro_str, proj=None):
    """
    normalize --distro argument as used by CLI

    set distro aliases from project if supplied (optional)

    return current distro if not set
    """
    if not distro_str:
        # current distro by default
        d = current_distro()
    elif isinstance(distro_str, Distro):
        # already correct format
        d = distro_str
    else:
        # convert custom distro string to Distro
        d = Distro(distro_str)
    if proj:
        d.aliases = proj.distro_aliases
    return d


def distro2idver(distro_name):
    """
    convert generic distro string into idver format
    """
    return re.sub(r'\s+', '-', distro_name.strip().lower())


def parse_idver(idver_str):
    """
    split idver string 'distro-version' into (distro, version)
    """
    id_, _, ver_ = idver_str.partition('-')
    return id_, ver_


def current_idver(sep='-'):
    """
    return current distro in default idver format
    used by apkg to reference individual distros

    examples: debian-10, fedora-32, arch
    """
    idver = distro.id()
    ver = distro.version()
    if ver:
        try:
            # check version string returned by "distro" is valid before appending
            version.parse(ver)
            idver += "%s%s" % (sep, ver)
        except version.InvalidVersion:
            pass
    return idver


def current_fullname():
    """
    return human readable string describing current distro
    """
    parts = [
        distro.name(),
        distro.version(pretty=True),
    ]
    return " ".join([p for p in parts if p])


def current_distro():
    return Distro(current_idver(sep=' '))



"""
Debian series custom template variables module for apkg.

Parses /usr/share/distro-info/$DISTRO.csv (from distro-info-data package)
to extract distro series (deb_series) and codename (deb_codename).

When no info is found for distro, return "unstable" / "Sid".

Provides custom template variables:

    deb_series : str    (examples: 'bookworm', 'noble', 'unstable')
    deb_codename : str  (examples: 'Bookworm', 'Noble Numbat', 'Sid')

Enable by adding [[template.variables]] entry to distro/config/apkg.toml:

    [[template.variables]]
    python_module = "apkg.templatevars.deb_series"

Example usage in debian/chanelog:

    some-package ({{ version }}-{{ release }}~{{ deb_series }}) {{ deb_series }}; urgency=medium

    * upstream package version {{ version }} for {{ distro }} {{ deb_codename }}

    -- Jakub Ružička <jru@debian.org>  {{ now }}
"""  # noqa

import csv
from collections import namedtuple


def get_distro_info(distro):
    distro_info_file = '/usr/share/distro-info/%s.csv' % distro
    DistroInfoTuple = None
    info = {}
    try:
        with open(distro_info_file, 'r') as csvfile:
            csvreader = csv.reader(csvfile)
            fields = next(csvreader)
            if not DistroInfoTuple:
                fields = list(map(lambda x: x.replace('-', '_'), fields))
                n_fields = len(fields)
                DistroInfoTuple = namedtuple('DistroInfoTuple', fields)

            for row in csvreader:
                while len(row) < n_fields:
                    row.append(None)
                dit = DistroInfoTuple(*row)
                info[dit.series] = dit
    except Exception:
        return {}

    return info


# called by apkg to get custom variables
def get_variables(env):
    distro_id = env['distro'].id
    distro_version = env['distro'].version
    if distro_version:
        distro_info = get_distro_info(distro_id)
        series_info = None
        for di in distro_info.values():
            # strip LTS and such
            version, _, _ = di.version.partition(' ')
            if distro_id != 'ubuntu':
                # strip minor releases for non-ubuntu (9.2 -> 9)
                version, _, _ = version.partition('.')
            if version == distro_version:
                series_info = di
                break

        if series_info:
            return {
                'deb_series': series_info.series,
                'deb_codename': series_info.codename,
            }

    return {
        'deb_series': 'unstable',
        'deb_codename': 'Sid',
    }



import glob
import hashlib
from pathlib import Path
import re

import jinja2
try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property

from apkg import adistro
from apkg import cache as _cache
from apkg import compat
from apkg import ex
from apkg.log import getLogger
from apkg import pkgtemplate
from apkg import pkgtest
from apkg.util.archive import unpack_archive
from apkg.util.git import git
from apkg.util import toml
from apkg.util import upstreamversion


log = getLogger(__name__)


INPUT_BASE_DIR = 'distro'
OUTPUT_BASE_DIR = 'pkg'
CONFIG_FN = 'apkg.toml'


# pylint: disable=too-many-instance-attributes
class ProjectPaths:
    # pylint: disable=redefined-builtin
    def __init__(self, base=None, input=None, output=None):
        self.base = Path(base or '.')
        if input:
            self.input = Path(input)
        else:
            self.input = self.base / INPUT_BASE_DIR
        if output:
            self.output = Path(output)
        else:
            self.output = self.base / OUTPUT_BASE_DIR
        self.update_paths()

    def update_paths(self):
        # project config: distro/config/apkg.toml
        self.config_base = self.input / 'config'
        self.config = self.config_base / CONFIG_FN
        # package templates: distro/pkg
        self.templates = self.input / 'pkg'
        # archives: pkg/archives/{dev,upstream,unpacked}
        self.archive = self.output / 'archives'
        self.dev_archive = self.archive / 'dev'
        self.upstream_archive = self.archive / 'upstream'
        self.unpacked_archive = self.archive / 'unpacked'
        # build: pkg/build/{src-,}pkg
        self.build = self.output / 'build'
        self.package_build = self.build / 'pkgs'
        self.srcpkg_build = self.build / 'srcpkgs'
        # output: pkg/{src-,}pkg
        self.package_out = self.output / 'pkgs'
        self.srcpkg_out = self.output / 'srcpkgs'
        # packaging tests: distro/tests
        self.tests = self.input / 'tests'
        # packaging tests extras: distro/tests/extra
        self.tests_extras = self.tests / 'extra'
        # packaging tests render/run: pkg/tests
        self.tests_out = self.output / 'tests'
        # cache: pkg/.cache.json
        self.cache = self.output / '.cache.json'


class Project:
    """
    Project class serves as high level interface to projecs in need of
    packaging
    """
    name = None
    config = {}
    distro_aliases = {}
    variables_sources = []

    def __init__(
            self,
            base_path=None,
            input_path=None,
            output_path=None,
            auto_load=True,
            auto_compat=True):
        self.auto_compat = auto_compat
        self.cache = _cache.ProjectCache(self)
        self.load_paths(
            base=base_path,
            input=input_path,
            output=output_path)
        if auto_load:
            self.load()

    def load(self):
        """
        load project config and update attributes
        """
        self.load_config()
        self.update_attrs()
        self.update_distro_aliases()
        self.update_variables_sources()
        if self.auto_compat:
            self.ensure_compat()

    def reload(self,
               base_path=None,
               input_path=None,
               output_path=None):
        """
        reload project with new paths supplied
        """
        self.load_paths(
            base=base_path or self.path.base,
            input=input_path or self.path.input,
            output=output_path or self.path.output)
        self.load()

    def load_paths(self, **kwargs):
        self.path = ProjectPaths(**kwargs)

    def load_config(self):
        """
        load project config from file
        """
        if self.path.config.exists():
            log.verbose("loading project config: %s", self.path.config)
            self.config = toml.loadp(self.path.config)
            return True
        else:
            log.verbose("project config not found: %s", self.path.config)
            return False

    def config_get(self, option, default=None):
        """
        get config option if set or default

        example options: 'project.name', 'upstream.archive_url'
        """
        c = self.config
        for key in option.split('.'):
            try:
                c = c[key]
            except KeyError:
                return default
        return c

    def update_attrs(self):
        """
        update project attributes based on current config
        """
        self.name = self.config_get('project.name')
        if self.name:
            log.verbose("project name from config: %s", self.name)
        else:
            self.name = self.path.base.resolve().name
            log.verbose("project name not in config - "
                        "guessing from path: %s", self.name)

    def update_distro_aliases(self):
        """
        load distro aliases from project config
        """
        conf = self.config_get('distro.aliases', [])
        self.distro_aliases = adistro.parse_distro_aliases(conf)

    def update_variables_sources(self):
        """
        load custom variables sources from project config
        """
        conf = self.config_get('template.variables', [])
        self.variables_sources = pkgtemplate.parse_variables_sources(conf)

    @cached_property
    def vcs(self):
        """
        Version Control System used in project

        possible outputs: 'git', None
        """
        o = git('rev-parse', quiet=True, check=False)
        if o.returncode == 0:
            return 'git'
        return None

    @cached_property
    def checksum(self):
        """
        checksum of current project state

        requires VCS (git), only computed once
        """
        if self.vcs == 'git':
            checksum = git.current_commit()[:10]
            diff = git('diff', quiet=True)
            if diff:
                diff_hash = hashlib.sha256(diff.encode('utf-8'))
                checksum += '-%s' % diff_hash.hexdigest()[:10]
            return checksum
        return None

    @property
    def compat_level(self):
        """
        current project compat level as set in config
        """
        return self.config_get('apkg.compat')

    def ensure_compat(self):
        compat.ensure_compat(self.compat_level)

    def upstream_archive_url(self, version):
        url = self.config_get('upstream.archive_url')
        if not url:
            return None
        tvars = {'project': self, 'version': version}
        url = jinja2.Template(url).render(**tvars)
        return url

    def upstream_signature_url(self, version):
        url = self.config_get('upstream.signature_url')
        if not url:
            return None
        tvars = {'project': self, 'version': version}
        url = jinja2.Template(url).render(**tvars)
        return url

    @cached_property
    def upstream_version(self):
        """
        check latest upstream version

        upstream is only queried once

        possible outputs: version, None
        """
        uv_script = self.config_get('upstream.version_script')
        if uv_script:
            v = upstreamversion.version_from_script(
                uv_script, script_name='upstream.version_script', proj=self)
            log.info("detected upstream version (from script): %s", v)
            return v
        ar_url = self.upstream_archive_url('VERSION')
        if ar_url:
            m = re.match(r'(.*/)[^/]+', ar_url)
            ar_base_url = m.group(1)
            v = upstreamversion.version_from_listing(ar_base_url)
            log.info("detected upstream version: %s", v)
            return v
        return None

    @cached_property
    def templates(self):
        if self.path.templates.exists():
            ignore_files = self.config_get('template.ignore_files')
            plain_copy_files = self.config_get('template.plain_copy_files')
            return pkgtemplate.load_templates(
                self.path.templates,
                distro_aliases=self.distro_aliases,
                ignore_files=ignore_files,
                plain_copy_files=plain_copy_files,
                variables_sources=self.variables_sources)
        else:
            return []

    def get_template_for_distro_(self, distro):
        for t in self.templates:
            if t.match_distro(distro):
                return t
        return None

    def get_template_for_distro(self, distro):
        if not isinstance(distro, adistro.Distro):
            distro = adistro.Distro(distro)
        template = self.get_template_for_distro_(distro)
        if not template:
            tdir = self.path.templates
            msg = ("missing package template for distro: %s\n\n"
                   "you can add it into: %s" % (distro.idver, tdir))
            raise ex.MissingPackagingTemplate(msg=msg)
        return template

    def load_upstream_archive(self, ar_path):
        """
        load project config from --upstream archive
        """
        unpack_path = unpack_archive(ar_path, self.path.unpacked_archive)
        input_path = unpack_path / 'distro'
        log.info("reloading project from upstream archive: %s", unpack_path)
        # reload project with input_path from archive
        self.reload(input_path=input_path)

    @cached_property
    def tests_extras(self):
        if self.path.tests_extras.exists():
            return pkgtest.load_tests_extras(
                self.path.tests_extras,
                distro_aliases=self.distro_aliases)
        else:
            return []

    def get_tests_extra_for_distro(self, distro):
        for e in self.tests_extras:
            if e.match_distro(distro):
                return e
        return None

    def get_tests_for_distro(self, distro):
        extra = self.get_tests_extra_for_distro(distro)
        tests = pkgtest.PackageTests(
            tests_in_path=self.path.tests,
            tests_out_path=self.path.tests_out,
            extra=extra,
            distro=distro)
        return tests

    def find_archives_by_name(self, name, upstream=False):
        """
        find archive files with supplied name in expected project paths
        """
        if upstream:
            ar_path = self.path.upstream_archive
        else:
            ar_path = self.path.dev_archive
        return glob.glob("%s/%s*" % (ar_path, name))



"""
apkg package style for **Debian**
and its many clones such as Ubuntu or Mint.

**source template**: content of `debian/` dir (`control`, `changelog`, ...)

**source package:** `*.dsc` + archives

**packages:** `*.deb`

**required distro packages**:

 * core: `devscripts`
 * isolated build: `pbuilder`

**template variables:**

 * `now`: current date in Debian changelog format (RFC 2822),
          use `SOURCE_DATE_EPOCH` env var when set
"""
import email
import email.utils
import glob
import os
from pathlib import Path
import re
import shutil
import tempfile
import time

from apkg import ex
from apkg.log import getLogger
from apkg import parse
from apkg.util.run import cd, run, sudo
from apkg.util.archive import unpack_archive


log = getLogger(__name__)


SUPPORTED_DISTROS = [
    "debian",
    "linuxmint",
    "pop",
    "raspbian",
    "ubuntu",
]
DISTRO_REQUIRES = {
    'core': ['build-essential'],
    'isolated': ['pbuilder'],
    'lint': ['lintian'],
}


def get_now():
    """
    current date and time in Debian changelog format (RFC 2822)

    SOURCE_DATE_EPOCH env variable overrides the time when set
    """
    t = int(os.environ.get('SOURCE_DATE_EPOCH', time.time()))
    return email.utils.formatdate(timeval=t, localtime=True)


TEMPLATE_VARS_DYNAMIC = {
    'now': get_now,
}


RE_PKG_NAME = r'Source:\s*(\S+)'
# orbital regexp cannon to parse Build-Depends from debian/control
RE_BUILD_DEPENDS = (
    r'(?:\n|\A)Build-Depends(?:-Indep)?:[ \t]*'  # no whitespace before
    r'(?:\n[ \t]+)?'  # optional leading newline with whitespace
    r'((?:[^,\n]+)'   # first build dep
    r'(?:,(?:[ \t]*'  # comma separator and optional whitespace
    r'(?:\n[ \t]+)?'  # optional newline starting with whitespace
    r'[^,\n]+))*)'    # 0-N other build deps
)


def is_valid_template(path):
    deb_files = ['rules', 'control', 'changelog']
    return all((path / f).exists() for f in deb_files)


def get_template_name(template, distro=None):
    """
    get Source package name from control
    """
    control_text = render_control_from_template_(template, distro=distro)
    for line in control_text.splitlines():
        m = re.match(RE_PKG_NAME, line)
        if m:
            return m.group(1)

    raise ex.ParsingFailed(
        msg="unable to determine Source from template: %s" % template.path)


def get_srcpkg_nvr(path):
    nvr, _, _ = str(path.name).rpartition('.')
    return nvr


def copy_pkg_files(src_path, dst_path, source=False):
    """
    copy package files at src_path to dst_path

    Find a single *.changes file at src_path
    and copy it and all files it references to dst_path.

    If multiple *.changes files are detected, select single *_source.changes if
    source=True, otherwise a single *_$ARCH.changes.

    Return a list of all files copied.

    This isn't a part of pkgstyle interface as it's fairly specific to Debian
    """
    changes = list(glob.iglob('%s/*.changes' % (src_path)))
    if not changes:
        raise ex.UnexpectedCommandOutput(
            msg="no *.changes files found when copying package")
    if len(changes) > 1:
        if source:
            # only select *_source.changes
            changes = [ch for ch in changes if ch.endswith('source.changes')]
        else:
            # only select *_$ARCH.changes
            changes = [ch for ch in changes if not ch.endswith('source.changes')]
        if len(changes) == 1:
            changes_fn = Path(changes[0]).name
            log.info("multiple *.changes files found, using %s changes: %s",
                     'source' if source else 'binary', changes_fn)
        else:
            raise ex.UnexpectedCommandOutput(
                msg="multiple *.changes files found when copying package")

    changes_path = Path(changes[0])
    changes_dst_path = dst_path / changes_path.name
    result = []
    shutil.copyfile(changes_path, changes_dst_path)
    result.append(changes_dst_path)

    parser = email.parser.HeaderParser()
    parsed = parser.parsestr(changes_path.read_text())
    for entry in parsed['Files'].split('\n'):
        if not entry:
            # first line is empty
            continue
        items = re.split(r'\s+', entry.strip(), maxsplit=5)
        if len(items) < 5:
            log.warning("Invalid *.changes line: %s", entry)
            continue
        name = items[4]
        src = src_path / name
        dst = dst_path / name
        shutil.copyfile(src, dst)
        result.append(dst)

    return result


def build_srcpkg(
        build_path,
        out_path,
        archive_info,
        template,
        tvars):
    """
    build debian source package
    """
    archive_path = archive_info["archive"]
    nv, _ = parse.split_archive_ext(archive_path.name)
    log.info("building deb source package: %s", nv)
    log.info("unpacking archive: %s", archive_path)
    source_path = unpack_archive(archive_path, build_path)
    log.verbose("source package root dir: %s", source_path)
    if not source_path or not source_path.exists():
        msg = "archive unpack didn't result in expected dir: %s" % source_path
        raise ex.UnexpectedCommandOutput(msg=msg)

    # copy archive with debian .orig name
    _, _, _, ext = parse.split_archive_fn(archive_path.name)
    debian_ar = "%s_%s.orig%s" % (tvars['name'], tvars['version'], ext)
    debian_ar_path = build_path / debian_ar
    log.info("copying archive into source package: %s", debian_ar_path)
    shutil.copyfile(archive_path, debian_ar_path)

    # extract components if any
    components = archive_info.get("components", {})
    for component, component_archive in components.items():
        log.info("unpacking component archive: %s", component)

        component_path = source_path / component
        # dpkg-source removes <component>/ and extracts the archive there
        if component_path.exists():
            shutil.rmtree(component_path)

        component_path.mkdir(exist_ok=True)
        # don't use unpack_archive from util.archive, component archives are
        # actually expected to be flat
        shutil.unpack_archive(component_archive, component_path)

        # copy archive with debian .orig-<component> name
        _, ext = parse.split_archive_ext(component_archive.name)

        debian_ar = "%s_%s.orig-%s%s" % (tvars['name'], tvars['version'],
                                         component, ext)
        debian_ar_path = build_path / debian_ar
        log.info("copying component archive into source package: %s",
                 debian_ar_path)
        shutil.copyfile(component_archive, debian_ar_path)

    # render template, dpkg-source removes the directory from upstream archive
    # first
    debian_path = source_path / 'debian'
    if debian_path.exists():
        shutil.rmtree(debian_path)
    template.render(debian_path, tvars=tvars)

    log.info("building deb source-only package...")
    with cd(source_path):
        run('dpkg-buildpackage',
            '-S',   # source-only, no binary files
            '-sa',  # source includes orig, always
            '-d',   # do not check build dependencies and conflicts
            '-nc',  # do not pre clean source tree
            '-us',  # unsigned source package.
            '-uc',  # unsigned .changes file.
            )

    log.info("copying source package to result dir: %s", out_path)
    out_path.mkdir(parents=True)
    copied = copy_pkg_files(build_path, out_path, source=True)
    dscs = [path for path in copied if path.suffix == '.dsc']
    if not dscs:
        raise ex.UnexpectedCommandOutput(
            msg="no *.dsc files found after source package build")
    return dscs


def build_packages(
        build_path,
        out_path,
        srcpkg_paths,
        **kwargs):
    """
    build .deb packages from source package
    """
    srcpkg_path = srcpkg_paths[0]
    build_path.mkdir(parents=True)
    out_path.mkdir(parents=True)
    isolated = kwargs.get('isolated')
    if isolated:
        log.info("starting isolated build using pbuilder")
        # TODO: ensure pbuilder's base image exists (pbuilder create)
        sudo('pbuilder', 'build',
             '--buildresult', build_path,
             srcpkg_path,
             preserve_env=True,  # preserve env inc. DEB_BUILD_OPTIONS
             )
    else:
        # unpack source package
        log.info("unpacking source package for direct host build")
        srcpkg_abspath = srcpkg_path.resolve()
        with cd(build_path):
            run('dpkg-source', '-x', srcpkg_abspath)
        # find unpacked source dir
        try:
            source_glob = '%s/*/' % build_path
            source_path = Path(glob.glob(source_glob)[0])
        except IndexError:
            msg = "failed to find unpacked source dir: %s"
            raise ex.UnexpectedCommandOutput(msg % source_glob)

        log.info("starting direct host build using dpkg-buildpackage")
        with cd(source_path):
            # build
            run('dpkg-buildpackage',
                '-sa',  # include orig, always
                '-us',  # unsigned source package.
                '-uc',  # unsigned .changes file.
                )

    log.info("copying built packages to result dir: %s", out_path)
    copied = copy_pkg_files(build_path, out_path, source=False)
    pkgs = [path for path in copied if path.suffix in ['.deb', '.ddeb']]
    if not pkgs:
        raise ex.UnexpectedCommandOutput(
            msg="no *.deb packages found after build")
    return pkgs


def install_distro_packages(
        packages,
        **kwargs):
    interactive = kwargs.get('interactive', False)
    reinstall = kwargs.get('reinstall', False)
    force = kwargs.get('force', False)

    cmd = ['apt-get']
    if reinstall:
        cmd += ['reinstall']
    else:
        cmd += ['install']
    env = os.environ.copy()
    if not interactive:
        env['DEBIAN_FRONTEND'] = 'noninteractive'
        cmd += ['-y']
    if force:
        cmd += ['--allow-downgrades']

    cmd += packages
    sudo(cmd, env=env)


def install_custom_packages(
        packages,
        **kwargs):

    def local_path(pkg):
        """
        apt-get is able to install local packages
        as long as they use full path or relative including ./
        """
        p = str(pkg)
        if p[0] not in '/\\.':
            return "./%s" % p
        return p

    interactive = kwargs.get('interactive', False)
    reinstall = kwargs.get('reinstall', False)
    force = kwargs.get('force', False)

    cmd = ['apt-get']
    if reinstall:
        cmd += ['reinstall']
    else:
        cmd += ['install']
    env = os.environ.copy()
    if not interactive:
        env['DEBIAN_FRONTEND'] = 'noninteractive'
        cmd += ['-y']
    if force:
        cmd += ['--allow-downgrades']

    cmd += list(map(local_path, packages))
    sudo(cmd, env=env)


def install_build_deps(
        deps,
        **kwargs):
    """
    install debian build deps

    Debian Build-Depends can contain strings not handled by
    `apt-get install` such as "(>= 9~)"

    New `apt-get satisfy` command handles Build-Depends strings fine
    but it isn't available on current.

    Try to use `apt-get satisfy` if available,
    otherwise revert to stripping special strings and use `install`.
    """
    interactive = kwargs.get('interactive', False)

    if has_aptget_satisfy_():
        # unlike install, satisfy can handle versioned deps
        cmd = ['apt-get', 'satisfy']
        env = os.environ.copy()
        if not interactive:
            env['DEBIAN_FRONTEND'] = 'noninteractive'
            cmd += ['-y']
        cmd += deps
        sudo(cmd, env=env)
    else:
        # satisfy not available, strip special strings and use install
        packages = [strip_dep_(d) for d in deps]
        install_distro_packages(packages, **kwargs)


def get_build_deps_from_template(
        template,
        distro=None):
    """
    parse Build-Depends from packaging template
    """
    control_text = render_control_from_template_(template, distro=distro)
    return get_build_deps_from_control_(control_text)


def render_control_from_template_(template, distro=None):
    """
    render control file from template
    """
    tvars = {}
    if distro:
        tvars['distro'] = distro
    control_text = template.render_file_content('control', tvars=tvars)
    return control_text


def get_build_deps_from_srcpkg(
        srcpkg_path,
        **_):
    """
    parse Build-Depends from source package
    """
    debar_path = get_srcpkg_debian_archive_(srcpkg_path.parent)
    log.info("unpacking debian archive: %s", debar_path)
    with tempfile.TemporaryDirectory(prefix='apkg_deb_') as td:
        unpack_path = unpack_archive(debar_path, td)
        control_path = unpack_path / 'control'
        control_text = control_path.open().read()
    return get_build_deps_from_control_(control_text)


def lint(
        pkg_paths,
        pedantic=False,
        info=False,
        strict=False,
        **kwargs):
    """
    lint files using lintian
    """
    log.info('linting %s files with lintian', len(pkg_paths))
    cmd = ['lintian']
    if pedantic:
        cmd += ['--display-info', '--pedantic']
    if info:
        cmd += ['--info']
    if strict:
        cmd += ['--fail-on', 'pedantic']
    cmd += pkg_paths
    o = run(cmd, check=False, direct=True)
    return o.returncode


# functions bellow with _ postfix are specific to this pkgstyle


def get_build_deps_from_control_(control_text):
    """
    parse Build-Depends from debian control file contents
    """
    m = re.findall(RE_BUILD_DEPENDS, control_text)
    if not m:
        msg = "unable to parse Build-Depends from control"
        raise ex.ParsingFailed(msg=msg)
    deps = []
    for deps_raw in m:
        deps += re.split(r'\s*,\s*', deps_raw)
    return deps


def get_srcpkg_debian_archive_(path):
    ars = glob.glob('%s/*.debian.tar.?z' % path)
    if not ars:
        msg = "unable to find debian archive in srcpkg: %s" % path
        raise ex.InvalidInput(msg=msg)
    if len(ars) > 1:
        msg = "multiple debian archives found in srcpkg: %s" % path
        raise ex.InvalidInput(msg=msg)
    return ars[0]


def has_aptget_satisfy_():
    """
    is `apt-get satisfy` command available?
    """
    o = run('apt-get', '-h', check=False, quiet=True)
    return 'satisfy' in o


def strip_dep_(dep):
    """
    strip special version strings (as found in Build-Depends)
    in order for dep to be installable through `apt-get install`
    """
    return re.split(r'[\s\[\(]', dep)[0]



class NullCallableString(str):
    """
    This emulates blessed class for cases when blessed isn't available
    """
    def __new__(cls):
        return str.__new__(cls, '')

    def __call__(self, arg, *extra_args):
        if isinstance(arg, int):
            return ''
        return arg


class PlainTerminal:
    """
    Mock of blessed Terminal that ignores formatting
    """
    nullstr = NullCallableString()

    def __getattr__(self, attr):
        setattr(self, attr, self.nullstr)
        return self.nullstr


COLOR_TERMINAL = False
COLOR_LIB = None
try:
    try:
        import blessed
        COLOR_LIB = 'blessed'
    except Exception:
        import blessings as blessed
        COLOR_LIB = 'blessings'
    # this can throw _curses.error: setupterm: could not find terminal
    # better find out now
    blessed.Terminal()
    Terminal = blessed.Terminal
    COLOR_TERMINAL = True
except Exception:
    Terminal = PlainTerminal



from pathlib import Path
import re
import requests

import click

from apkg import ex
from apkg.commands.make_archive import archive_dict2list
from apkg.util import common
from apkg.log import getLogger
from apkg.project import Project


log = getLogger(__name__)


@click.command(name='get-archive')
@click.option('-v', '--version',
              help='version of archive to download')
@click.option('-O', '--result-dir',
              help="put results into specified dir")
@click.option('-o', '--out-format', default='yaml', show_default=True,
              type=click.Choice(['yaml', 'list']),
              help="set output format")
@click.option('--cache/--no-cache', default=True, show_default=True,
              help="enable/distable cache")
@click.help_option('-h', '--help', help='show this help')
def cli_get_archive(*args, **kwargs):
    """
    download upstream archive for current project
    """
    out_format = kwargs.pop('out_format')

    results = get_archive(*args, **kwargs)

    if out_format == 'list':
        rlist = archive_dict2list(results)
        common.print_results(rlist)
    else:
        # YAML
        common.print_results_dict(results)

    return results


def get_archive(
        version=None,
        result_dir=None,
        cache=True,
        project=None):
    """
    download upstream archive for current project

    upstream.archive_url config option specifies archive to download.

    If upstream.signature_url config option is provided,
    target signature file is downloaded alongside the archive.

    When no --version is specified, apkg tries to detect latest version:

    1) using upstream.version_script if set
    2) from HTML listing if upstream.archive_url is set
    """
    proj = project or Project()
    if not version:
        version = proj.upstream_version
        if not version:
            raise ex.UnableToDetectUpstreamVersion()
    archive_url = proj.upstream_archive_url(version)

    use_cache = proj.cache.enabled(
        'remote', cmd='get_archive', use_cache=cache)
    if use_cache:
        cache_key = 'archive/upstream/%s' % archive_url
        cached = common.get_cached_paths(proj, cache_key, result_dir)
        if cached:
            log.success("reuse cached archive: %s", cached['archive'])
            return cached

    log.info('downloading archive: %s', archive_url)

    _, _, archive_fn = archive_url.rpartition('/')
    if result_dir:
        ar_base_path = Path(result_dir)
    else:
        ar_base_path = proj.path.upstream_archive

    r = requests.get(archive_url, allow_redirects=True)
    if r.status_code != 200:
        raise ex.FileDownloadFailed(code=r.status_code, url=archive_url)
    content_type = r.headers.get('content-type', '')
    if not content_type.startswith('application/'):
        msg = 'Failed to download archive - invalid content-type "%s":\n\n%s'
        raise ex.FileDownloadFailed(
            msg=msg % (content_type, archive_url))

    content_disp = r.headers.get('content-disposition')
    if content_disp:
        m = re.search(r'filename=(.+)', content_disp)
        if m:
            archive_fn = m.group(1).strip(' "')
            log.verbose("archive file name from HTTP Content-Disposition: %s",
                        archive_fn)

    archive_path = ar_base_path / archive_fn
    log.info('saving archive to: %s', archive_path)
    ar_base_path.mkdir(parents=True, exist_ok=True)
    archive_path.open('wb').write(r.content)
    log.success('downloaded archive: %s', archive_path)

    results = {
        'archive': archive_path,
        'archive_url': archive_url,
        'version': str(version),
    }

    signature_url = proj.upstream_signature_url(version)
    if signature_url:
        # singature check
        log.info('downloading signature: %s', signature_url)
        r = requests.get(signature_url, allow_redirects=True)
        if not r.ok:
            raise ex.FileDownloadFailed(
                code=r.status_code, url=signature_url)
        _, _, signature_name = signature_url.rpartition('/')
        signature_path = ar_base_path / signature_name
        log.info('saving signature to: %s', signature_path)
        signature_path.open('wb').write(r.content)
        log.success('downloaded signature: %s', signature_path)
        results['signature'] = signature_path
        results['signature_url'] = signature_url
    else:
        log.verbose("project.upstream_signature_url not set"
                    " - skipping signature download")

    if use_cache:
        proj.cache.update(cache_key, results)

    return results


def parse_archive_args(proj, archive, upstream, inputs):
    """
    Helper to parse archive arguments,
    use get-archive,
    and reload project config from archive
    as appropriate.

    Useful for other commands.
    """
    archive_files = inputs
    if upstream:
        archive = True
        if not archive_files:
            archive_files = get_archive(project=proj)
    if archive:
        common.ensure_inputs(archive_files)
        proj.load_upstream_archive(archive_files[0])
    return archive, archive_files


APKG_CLI_COMMANDS = [cli_get_archive]



import click
import shutil

from apkg import ex
from apkg.log import getLogger
from apkg.project import Project
from apkg.util.git import git


log = getLogger(__name__)


@click.command(name='clean')
@click.option('-c', '--cache', is_flag=True,
              help="clean apkg cache metadata only")
@click.option('--hard', is_flag=True,
              help=("[!] HARD RESET project from VCS "
                    "and REMOVE extra files"))
@click.help_option('-h', '--help', help='show this help')
def cli_clean(*args, **kwargs):
    """
    clean apkg output directory (pkg/)
    """
    clean(*args, **kwargs)


def clean(cache=False, hard=False):
    proj = Project()
    if hard:
        if proj.vcs != 'git':
            raise ex.InvalidUsage(
                msg="HARD RESET currently requires git, "
                "but project VCS is: %s" % proj.vcs)
        git('reset', '--hard')
        git('clean', '-fxd')
        return

    if cache:
        log.bold("cleaning apkg cache")
        if proj.path.cache.exists():
            proj.path.cache.unlink()
            log.success("removed apkg cache: %s", proj.path.cache)
        else:
            log.success("apkg cache doesn't exist: %s", proj.path.cache)
        return

    log.bold("cleaning apkg output")
    if proj.path.output.exists():
        shutil.rmtree(proj.path.output)
        log.success("removed apkg output dir: %s", proj.path.output)
    else:
        log.success("apkg output dir doesn't exist: %s", proj.path.output)


APKG_CLI_COMMANDS = [cli_clean]



import click

from apkg import adistro
from apkg.cli import cli
from apkg.pkgstyle import call_pkgstyle_fun
from apkg.commands.get_archive import parse_archive_args
from apkg.commands.srcpkg import srcpkg as cmd_srcpkg
from apkg.commands.test_dep import test_dep as cmd_test_dep
from apkg.util import common
from apkg.log import getLogger
from apkg.project import Project


log = getLogger(__name__)


@cli.command(name="build-dep", aliases=['builddep'])
@click.argument('inputs', nargs=-1)
@click.option('-l', '--list', 'list', is_flag=True,
              help="list build deps only, don't install")
@click.option('-t', '--test-dep', is_flag=True,
              help="include testing deps as well")
@click.option('-u', '--upstream', is_flag=True,
              help="use upstream template / archive / srcpkg")
@click.option('-s', '--srcpkg', is_flag=True,
              help="use source package")
@click.option('-a', '--archive', is_flag=True,
              help="use template (/build srcpkg) from archive")
@click.option('-d', '--distro',
              help="override target distro  [default: current]")
@click.option('-F', '--in-file', 'in_files', multiple=True,
              help="specify input file, '-' to read from stdin")
@click.option('--ask/--no-ask', 'interactive',
              default=False, show_default=True,
              help="enable/disable interactive mode")
# TODO: once py3.6 is dropped, add hidden=True
@click.option('-y', '--yes', 'interactive', flag_value=False,
              help="[DEPRECATED] compat alias for --no-ask")
@click.help_option('-h', '--help',
                   help="show this help message")
def cli_build_dep(*args, **kwargs):
    """
    install or list build dependencies
    """
    kwargs['install'] = not kwargs.pop('list')
    deps = build_dep(*args, **kwargs)
    if not kwargs['install']:
        common.print_results(deps)


def build_dep(
        test_dep=False,
        upstream=False,
        srcpkg=False,
        archive=False,
        inputs=None,
        in_files=None,
        install=True,
        distro=None,
        interactive=False,
        project=None):
    """
    parse and optionally install build dependencies

    pass install=False to only get list of deps without install

    returns list of build deps
    """
    action = 'installing' if install else 'listing'
    log.bold('%s build deps', action)

    proj = project or Project()
    distro = adistro.distro_arg(distro, proj)
    log.info("target distro: %s", distro)

    inputs = common.parse_inputs(inputs, in_files)

    if srcpkg:
        # use source package to determine deps
        if archive or not inputs:
            # build source package
            srcpkg_files = cmd_srcpkg(
                archive=archive,
                distro=distro,
                inputs=inputs,
                in_files=in_files,
                upstream=upstream,
                project=proj)
        else:
            # use specified source package
            srcpkg_files = inputs

        common.ensure_inputs(srcpkg_files)
        srcpkg_path = srcpkg_files[0]

        # fetch pkgstyle (deb, rpm, arch, ...)
        template = proj.get_template_for_distro(distro)
        pkgstyle = template.pkgstyle

        log.info("build deps from srcpkg: %s", srcpkg_path)
        deps = call_pkgstyle_fun(
            pkgstyle, 'get_build_deps_from_srcpkg',
            srcpkg_path)

        if test_dep:
            log.warning("unable to load test deps from --srcpkg - skipping")
    else:
        # use tempalte to determine deps
        archive, archive_files = parse_archive_args(
            proj, archive, upstream, inputs)

        # fetch pkgstyle (deb, rpm, arch, ...)
        template = proj.get_template_for_distro(distro)
        pkgstyle = template.pkgstyle

        log.info("build deps from template: %s", template.path)
        deps = call_pkgstyle_fun(
            pkgstyle, 'get_build_deps_from_template',
            template, distro=distro)

        if test_dep:
            # include test deps as well
            tdeps = cmd_test_dep(
                install=False,
                upstream=upstream,
                archive=archive,
                inputs=archive_files)
            deps = sorted(set(deps).union(tdeps))

    if install:
        n_deps = len(deps)
        if n_deps > 0:
            log.info("installing %s build deps...", n_deps)
            call_pkgstyle_fun(
                pkgstyle, 'install_build_deps',
                deps,
                distro=distro,
                interactive=interactive)
        else:
            log.info("no build deps to install")

    return deps


APKG_CLI_COMMANDS = [cli_build_dep]



import click

from apkg import adistro
from apkg.commands.build import build
from apkg import ex
from apkg.log import getLogger
from apkg import pkgstyle
from apkg.util import common


log = getLogger(__name__)


@click.command(name="install")
@click.argument('inputs', nargs=-1)
@click.option('-C', '--custom-pkgs', is_flag=True,
              help="install custom packages (no build)")
@click.option('-D', '--distro-pkgs', is_flag=True,
              help="install packages from distro repos (no build)")
@click.option('-s', '--srcpkg', is_flag=True,
              help="build and install packages from srcpkg")
@click.option('-a', '--archive', is_flag=True,
              help="build and install packages from archive")
@click.option('-u', '--upstream', is_flag=True,
              help="build and install upstream packages")
@click.option('-v', '--version',
              help=("upstream archive version to use"
                    ", implies --upstream"))
@click.option('-r', '--release',
              help="set release for built packages  [default: 1]")
@click.option('-b', '--build-dep', is_flag=True,
              help="install build deps on host (apkg build-dep)")
@click.option('-d', '--distro',
              help="override target distro  [default: current]")
@click.option('--reinstall', is_flag=True,
              help="[!] reinstall existing packages")
@click.option('--force', is_flag=True,
              help="[!] allow dangerous operations (downgrade, erase, replace)")
@click.option('--cache/--no-cache', default=True, show_default=True,
              help="enable/disable cache")
@click.option('--ask/--no-ask', 'interactive',
              default=False, show_default=True,
              help="enable/disable interactive mode")
@click.option('-F', '--in-file', 'in_files', multiple=True,
              help="specify input file, '-' to read from stdin")
# TODO: once py3.6 is dropped, add hidden=True
@click.option('-y', '--yes', 'interactive', flag_value=False,
              help="[DEPRECATED] compat alias for --no-ask")
@click.help_option('-h', '--help',
                   help="show this help message")
def cli_install(*args, **kwargs):
    """
    install packages using native package manager

    Default: build packages and install them

    `apkg build` options are available

    you can supply a list of --custom-pkgs or --distro-pkgs to install
    specified custom/distro packages directly without any build
    """
    return install(*args, **kwargs)


def install(
        custom_pkgs=False,
        distro_pkgs=False,
        srcpkg=False,
        archive=False,
        upstream=False,
        inputs=None,
        in_files=None,
        version=None,
        release=None,
        distro=None,
        build_dep=False,
        reinstall=False,
        force=False,
        cache=True,
        interactive=False):
    """
    install packages using native package manager
    """
    log.bold("installing packages")

    distro = adistro.distro_arg(distro)
    log.info("target distro: %s", distro)

    ps = pkgstyle.get_pkgstyle_for_distro(distro)
    if not ps:
        raise ex.DistroNotSupported(distro=distro)
    log.info("target pkgstyle: %s", ps.name)

    inputs = common.parse_inputs(inputs, in_files)

    pkgs = []
    result = None

    if custom_pkgs:
        if distro_pkgs:
            raise ex.InvalidInput(
                fail=("--custom-pkgs and --distro-pkgs options"
                      " are mutually exclusive"))

        pkgs = inputs
        result = pkgstyle.call_pkgstyle_fun(
            ps, 'install_custom_packages',
            pkgs,
            distro=distro,
            reinstall=reinstall,
            force=force,
            interactive=interactive)
    elif distro_pkgs:
        pkgs = inputs
        result = pkgstyle.call_pkgstyle_fun(
            ps, 'install_distro_packages',
            pkgs,
            distro=distro,
            reinstall=reinstall,
            force=force,
            interactive=interactive)
    else:
        # default: use build to get packages
        pkgs = build(
            srcpkg=srcpkg,
            archive=archive,
            upstream=upstream,
            inputs=inputs,
            version=version,
            release=release,
            distro=distro,
            build_dep=build_dep,
            cache=cache)

        result = pkgstyle.call_pkgstyle_fun(
            ps, 'install_custom_packages',
            pkgs,
            distro=distro,
            reinstall=reinstall,
            force=force,
            interactive=interactive)

    log.success("installed %s packages", len(pkgs))

    return result


APKG_CLI_COMMANDS = [cli_install]



"""
apkg package style for **Arch** linux.

**source template:** `PKGBUILD`

**source package:** `PKGBUILD`

**packages:** `*.zst`
"""
import glob
from pathlib import Path
import shutil

from apkg import ex
from apkg.log import getLogger
from apkg.util.run import cd, run, sudo


log = getLogger(__name__)


SUPPORTED_DISTROS = [
    'arch',
    'manjaro',
]
DISTRO_REQUIRES = {
    'core': ['base-devel'],
}


RE_PKG_NAME = r'pkgname\s*=\s*(\S+)'


def is_valid_template(path):
    pkgbuild = path / "PKGBUILD"
    return pkgbuild.exists()


def get_template_name(template, distro=None):
    pkgbuild_text = render_pkgbuild_from_template_(template, distro=distro)
    return parse_pkgbuild_content_(pkgbuild_text, 'echo "$pkgname"')


def get_srcpkg_nvr(path):
    return parse_pkgbuild_(path, 'echo "$pkgname-$pkgver-$pkgrel"')


def build_srcpkg(
        build_path,
        out_path,
        archive_info,
        template,
        tvars):
    archive_path = archive_info['archive']
    in_pkgbuild = build_path / 'PKGBUILD'
    out_pkgbuild = out_path / 'PKGBUILD'
    out_archive = out_path / archive_path.name
    log.info("building arch source package: %s", in_pkgbuild)
    template.render(build_path, tvars or {})
    out_path.mkdir(parents=True)

    log.info("copying PKGBUILD and archive to: %s", out_path)
    shutil.copyfile(in_pkgbuild, out_pkgbuild)
    archive_paths = [archive_info["archive"]]
    archive_paths.extend(archive_info.get("components", {}).values())
    for src_path in archive_paths:
        dst_path = out_path / src_path.name
        shutil.copyfile(src_path, dst_path)

    return [out_pkgbuild, out_archive]


def build_packages(
        build_path,
        out_path,
        srcpkg_paths,
        **_):
    srcpkg_path = srcpkg_paths[0]
    if srcpkg_path.name != 'PKGBUILD':
        raise ex.InvalidSourcePackageFormat(
            fmt="arch source package format is PKGBUILD but got: %s"
            % srcpkg_path.name)
    log.info("copying source package to build dir: %s", build_path)
    shutil.copytree(srcpkg_path.parent, build_path)
    # build package using makepkg
    log.info("starting arch package build using makepkg")
    with cd(build_path):
        run('makepkg')
    log.info("copying built packages to result dir: %s", out_path)
    out_path.mkdir(parents=True, exist_ok=True)
    pkgs = []
    # find and copy resulting packages
    for src_pkg in glob.iglob('%s/*.zst' % build_path):
        dst_pkg = out_path / Path(src_pkg).name
        shutil.copyfile(src_pkg, dst_pkg)
        pkgs.append(dst_pkg)

    return pkgs


def install_distro_packages(
        packages,
        **kwargs):
    interactive = kwargs.get('interactive', False)
    cmd = ['pacman', '-S', '--needed']
    if not interactive:
        cmd += ['--noconfirm']
    cmd += packages
    sudo(cmd)


def install_custom_packages(
        packages,
        **kwargs):
    interactive = kwargs.get('interactive', False)
    cmd = ['pacman', '-U']
    if not interactive:
        cmd += ['--noconfirm']
    cmd += packages
    sudo(cmd)


def install_build_deps(
        deps,
        **kwargs):
    # no special handling for build deps on arch
    install_distro_packages(deps, **kwargs)


def get_build_deps_from_template(
        template,
        distro=None):
    """
    parse depends and makedepends from packaging template
    """
    # render PKGBUILD
    pkgbuild_text = render_pkgbuild_from_template_(template, distro=distro)
    # depends: required for build and run
    depends = parse_pkgbuild_content_(
        pkgbuild_text,
        'printf \'%s\n\' "${depends[@]}"').splitlines()
    # makedepends: required for build only
    makedepends = parse_pkgbuild_content_(
        pkgbuild_text,
        'printf \'%s\n\' "${makedepends[@]}"').splitlines()
    return depends + makedepends


def render_pkgbuild_from_template_(template, distro=None):
    """
    render PKGBUILD file from template
    """
    tvars = {}
    if distro:
        tvars['distro'] = distro
    pkgbuild_txt = template.render_file_content('PKGBUILD', tvars=tvars)
    return pkgbuild_txt


def get_build_deps_from_srcpkg(
        srcpkg_path,
        **_):
    """
    parse depends from source package (PKGBUILD)
    """
    deps = parse_pkgbuild_(
        srcpkg_path,
        'printf \'%s\n\' "${depends[@]}"')
    return deps.splitlines()


# functions bellow with _ postfix are specific to this pkgstyle


def parse_pkgbuild_(pkgbuild, bash):
    return run('bash', '-c', '. "%s" && %s'
               % (pkgbuild, bash), quiet=True)


def parse_pkgbuild_content_(pkgbuild_text, bash):
    pkgb = '%s\n%s' % (pkgbuild_text, bash)
    return run('bash', '-s', input=pkgb, quiet=True)



import click

from apkg import adistro
from apkg.cli import cli
from apkg.pkgstyle import call_pkgstyle_fun, get_pkgstyle_for_distro
from apkg.commands.get_archive import parse_archive_args
from apkg.util import common
from apkg.log import getLogger
from apkg.project import Project


log = getLogger(__name__)


@cli.command(name="test-dep", aliases=['testdep'])
@click.argument('inputs', nargs=-1)
@click.option('-l', '--list', 'list', is_flag=True,
              help="list test deps only, don't install")
@click.option('-u', '--upstream', is_flag=True,
              help="use upstream archive")
@click.option('-a', '--archive', is_flag=True,
              help="use test deps from archive")
@click.option('-d', '--distro',
              help="override target distro  [default: current]")
@click.option('-F', '--in-file', 'in_files', multiple=True,
              help="specify input file, '-' to read from stdin")
@click.option('--ask/--no-ask', 'interactive',
              default=False, show_default=True,
              help="enable/disable interactive mode")
@click.help_option('-h', '--help',
                   help="show this help message")
def cli_test_dep(*args, **kwargs):
    """
    install or list testing dependencies
    """
    kwargs['install'] = not kwargs.pop('list')
    deps = test_dep(*args, **kwargs)
    if not kwargs['install']:
        common.print_results(deps)


def test_dep(
        upstream=False,
        archive=False,
        inputs=None,
        in_files=None,
        install=True,
        distro=None,
        interactive=False,
        project=None):
    """
    parse and optionally install testing dependencies

    pass install=False to only get list of deps without install

    returns list of test deps
    """
    action = 'installing' if install else 'listing'
    log.bold('%s testing deps', action)

    proj = project or Project()
    distro = adistro.distro_arg(distro, proj)
    log.info("target distro: %s", distro)

    inputs = common.parse_inputs(inputs, in_files)
    archive, _ = parse_archive_args(
        proj, archive, upstream, inputs)

    tests = proj.get_tests_for_distro(distro)
    deps = tests.deps

    if install:
        n_deps = len(deps)
        if n_deps > 0:
            log.info("installing %s test deps...", n_deps)
            pkgstyle = get_pkgstyle_for_distro(distro)
            call_pkgstyle_fun(
                pkgstyle, 'install_build_deps',
                deps,
                distro=distro,
                interactive=interactive)
        else:
            log.info("no build deps to install")

    return deps


APKG_CLI_COMMANDS = [cli_test_dep]



import click

from apkg import adistro
from apkg.cli import cli
from apkg.commands.get_archive import parse_archive_args
from apkg.commands.test_dep import test_dep as cmd_test_dep
from apkg import ex
from apkg.log import getLogger, T
from apkg.project import Project
from apkg.util import common


log = getLogger(__name__)


@cli.command(name="test", aliases=['tests'])
@click.argument('inputs', nargs=-1)
@click.option('-i', '--info', is_flag=True,
              help="show tests information and exit")
@click.option('-l', '--list-tests', is_flag=True,
              help="list tests and exit")
@click.option('-c', '--show-control', is_flag=True,
              help="print tests control file and exit (render if needed)")
@click.option('-u', '--upstream', is_flag=True,
              help="use upstream archive")
@click.option('-a', '--archive', is_flag=True,
              help="use test deps from archive")
@click.option('-d', '--distro',
              help="override target distro  [default: current]")
@click.option('-t', '--test-dep', is_flag=True,
              help="install testing dependencies on host (apkg test-dep)")
@click.option('-k', '--test-filter',
              help="only select tests matching supplied REGEX")
@click.option('-F', '--in-file', 'in_files', multiple=True,
              help="specify input file, '-' to read from stdin")
@click.help_option('-h', '--help',
                   help="show this help message")
def cli_test(*args, **kwargs):
    """
    run packaging tests
    """
    r = test(*args, **kwargs)
    if not r:
        raise ex.PkgTestFail


def test(
        info=False,
        list_tests=False,
        show_control=False,
        upstream=False,
        archive=None,
        inputs=None,
        in_files=None,
        distro=None,
        test_dep=False,
        test_filter=None):
    """
    run packaging tests
    """
    proj = Project()
    distro = adistro.distro_arg(distro, proj)
    inputs = common.parse_inputs(inputs, in_files)
    archive, archive_files = parse_archive_args(
        proj, archive, upstream, inputs)

    if test_dep:
        cmd_test_dep(
            upstream=upstream,
            archive=None,
            inputs=archive_files,
            distro=distro)

    if info:
        return test_info(project=proj, distro=distro)

    tests = proj.get_tests_for_distro(distro)

    if show_control:
        if not tests.control_path.exists():
            log.error('tests control file not found: %s', tests.control_path)
            return False
        if tests.extra_control:
            msg = 'extra tests control template: %s'
        else:
            msg = 'inline tests control file: %s'
        log.info(msg, tests.control_path)
        print(tests.control_text)
        return bool(tests.control_text)

    if list_tests:
        tests_ = tests.filter_tests(test_filter)
        for t in tests_:
            if t.name:
                tests_path = t.tests_dir or proj.path.tests
                cmd = str(tests_path / t.name)
            elif t.test_command:
                cmd = t.test_command
            print(cmd)
        return tests_

    log.bold("running packaging tests")
    log.info("target distro: %s", distro)

    return tests.run(test_filter=test_filter)


def test_info(project=None, distro=None):
    """
    show packaging tests info
    """
    proj = project or Project()
    distro = adistro.distro_arg(distro, proj)

    msg = "tests path:         {t.bold}{fn}{t.normal}"
    if proj.path.tests.exists():
        msg += " ({t.green}exists{t.normal})"
    else:
        msg += " ({t.error}doesn't exist{t.normal})"
    print(msg.format(fn=proj.path.tests, t=T))
    msg = "tests extras path:  {t.bold}{fn}{t.normal}"
    if proj.path.tests_extras.exists():
        msg += " ({t.green}exists{t.normal})"
    else:
        msg += " (doesn't exist)"
    print(msg.format(fn=proj.path.tests_extras, t=T))

    msg = "tests extras:"
    if proj.tests_extras:
        msg_lines = []
        for extra in proj.tests_extras:
            msg_lines.append(
                "    {t.bold}%s{t.normal}: %s: %s"
                % (extra.name, extra.selection_str(),
                   extra.distro_rules))
        msg = "\n".join([msg] + msg_lines)
    else:
        msg += "       {t.bold}N/A{t.normal}"
    print(msg.format(dir=proj.path.templates, t=T))

    print()

    # TODO: pass tvars to test control template
    tests = proj.get_tests_for_distro(distro)
    msg = "testing distro:     {t.cyan}{d}{t.normal}"
    print(msg.format(d=distro, t=T))
    msg = "tests control:      {t.bold}{p}{t.normal}"
    if tests.control_path.exists():
        msg += " ({t.green}exists{t.normal})"
    else:
        msg += " ({t.error}doesn't exist{t.normal})"
    print(msg.format(p=tests.control_path, t=T))
    msg = "tests extra:        {t.bold}{p}{t.normal}"
    print(msg.format(p=tests.extra_path or 'N/A', t=T))

    print()

    print("tests:")
    if tests.tests:
        for t in tests.tests:
            print("    {t.bold}{test}{t.normal}".format(test=t, t=T))
            if t.tests_dir:
                print("        Tests-Directory: %s" % t.tests_dir)
            if t.deps:
                print("        Depends: %s" % ", ".join(t.deps))
            if t.restrictions:
                print("        Restrictions: %s" % " ".join(t.restrictions))
    else:
        msg = "    {t.error}no tests found{t.error}"
        print(msg.format(t=T))

    return tests.tests


APKG_CLI_COMMANDS = [cli_test]



"""
apkg the cross-distro packaging automation tool
"""
import os
import pkgutil
import sys

import click

from apkg import __version__
from apkg import commands
from apkg import ex
from apkg.log import getLogger, T
from apkg.util import common
from apkg.util.run import log_cmd_fail
import apkg.log as _log


log = getLogger(__name__)


CLI_LOG_LEVELS = {
    'debug': _log.DEBUG,
    'verbose': _log.VERBOSE,
    'info': _log.INFO,
    'brief': _log.WARN,
    'quiet': _log.ERROR,
}
CLI_PATH_FORMATS = ['rel', 'abs', 'stem']


class ClickApkgGroup(click.Group):
    """
    custom click group for handling command aliases

    based on https://github.com/click-contrib/click-aliases
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._commands = {}
        self._aliases = {}

    def command(self, *args, **kwargs):
        aliases = kwargs.pop('aliases', [])
        decorator = super().command(*args, **kwargs)
        if not aliases:
            return decorator

        def _decorator(f):
            cmd = decorator(f)
            if aliases:
                self._commands[cmd.name] = aliases
                for alias in aliases:
                    self._aliases[alias] = cmd.name
            return cmd

        return _decorator

    def group(self, *args, **kwargs):
        aliases = kwargs.pop('aliases', [])
        decorator = super().group(*args, **kwargs)
        if not aliases:
            return decorator

        def _decorator(f):
            cmd = decorator(f)
            if aliases:
                self._commands[cmd.name] = aliases
                for alias in aliases:
                    self._aliases[alias] = cmd.name
            return cmd

        return _decorator

    def resolve_alias(self, cmd_name):
        if cmd_name in self._aliases:
            return self._aliases[cmd_name]
        return cmd_name

    def get_command(self, ctx, cmd_name):
        cmd_name = self.resolve_alias(cmd_name)
        command = super().get_command(ctx, cmd_name)
        return command


@click.group(cls=ClickApkgGroup)
@click.option('-L', '--log-level',
              default='info', show_default=True,
              type=click.Choice(CLI_LOG_LEVELS.keys()),
              help="set log level")
@click.option('-P', '--path-format',
              default='rel', show_default=True,
              type=click.Choice(CLI_PATH_FORMATS),
              help="set file paths format")
@click.help_option('-h', '--help',
                   help="show this help message")
@click.version_option(__version__, message='%(version)s',
                      help="show apkg version")
def cli(log_level='info', path_format='rel'):
    """
    apkg the upstream packaging automation tool
    """
    level = CLI_LOG_LEVELS[log_level]
    _log.set_log_level(level)
    log.verbose("apkg version: %s", __version__)
    log.verbose("log level: %s (%s)", log_level.upper(), _log.get_log_level())
    common.set_path_format(path_format)
    log.verbose("path format: %s", common.get_path_format())


def apkg(*args):
    """
    apkg shell interface

    Execute apkg command with specified arguments
    and return shell friendly exit code.

        py> apkg('command', 'argument')

    is equivalent to

        $> apkg command argument

    This is a shell-friendly wrapper around cli()
    which always returns numeric exit code
    as opposed to raising exceptions.

    If you're using apkg from python you probably want to use
    cli() or apkg.commands.* modules directly instead.
    """

    code = 1
    try:
        # pylint: disable=unexpected-keyword-arg
        cli(args, standalone_mode=False)
        code = 0
    except ex.CommandFailed as e:
        log_cmd_fail(e.kwargs['cmdout'])
        code = e.returncode
    except ex.PkgTestFail as e:
        # already logged
        code = e.returncode
    except ex.QuietExit as e:
        code = e.returncode
    except ex.ApkgException as e:
        print(T.bold_yellow(str(e)))
        code = e.returncode
    except click.exceptions.ClickException as e:
        print(T.yellow(str(e)))
        code = ex.InvalidUsage.returncode

    return code


def main():
    """
    apkg console_scripts entry point
    """
    cargs = sys.argv[1:]
    sys.exit(apkg(*cargs))


def __load_commands():
    """
    load available apkg commands

    should only be called once on module load
    """
    pkgpath = os.path.dirname(commands.__file__)
    for _, modname, _ in pkgutil.iter_modules([pkgpath]):
        modpath = "apkg.commands.%s" % (modname)
        mod = __import__(modpath, fromlist=[''])
        cmds = getattr(mod, 'APKG_CLI_COMMANDS', None)
        if not cmds:
            log.warning('command module with no CLI commands: %s', modpath)
            continue
        for cmd in cmds:
            cli.add_command(cmd)


__load_commands()


if __name__ == '__main__':
    main()



# -*- encoding: utf-8 -*-

import re

import bs4
from packaging import version
import requests
import yaml

from apkg import ex
from apkg.log import getLogger
from apkg.util.run import run


log = getLogger(__name__)


RE_ARCHIVE_VERSION = r'[\w-]+-(\d[^-]+)\.tar\..*'


def version_from_script(script, script_name='script', proj=None):
    """
    get version from a script
    """
    log.verbose("getting upstream version from %s: %s", script_name, script)
    out = run(script, quiet=True)
    if proj and proj.compat_level < 6:
        # use old behaviour - the last line of output was used
        msg = ("using legacy %s format (last line)"
               " due to compat level %s" % (script_name, proj.compat_level))
        log.info(msg)
        _, _, last_line = out.rpartition('\n')
        v = last_line.strip()
    else:
        # modern YAML output (compat >= 6)
        try:
            results = yaml.safe_load(out)
        except Exception as e:
            msg = ("Failed to parse %s YAML output:\n\n"
                   "%s\n\nError:\n\n%s" % (script_name, out, e))
            raise ex.UnexpectedCommandOutput(msg=msg)
        if not isinstance(results, dict) or 'version' not in results:
            msg = ("Invalid %s YAML output format:\n\n"
                   "%s\n\nExpected format example:\n\n"
                   "version: 1.2.3" % (script_name, out))
            raise ex.UnexpectedCommandOutput(msg=msg)
        v = results['version']
    v = version.parse(v)
    return v


def version_from_listing(html_listing_url):
    """
    get latest version from HTML listing
    """
    log.verbose("getting upstream version from HTML listing: %s",
                html_listing_url)
    found = False
    v_max = version.parse('0')
    r = requests.get(html_listing_url)
    soup = bs4.BeautifulSoup(r.content, 'html.parser')
    for a in soup.find_all('a'):
        m = re.match(RE_ARCHIVE_VERSION, a.string)
        if not m:
            continue
        v_str = m.group(1)
        try:
            v = version.parse(v_str)
        except version.InvalidVersion:
            log.verbose("Ignoring invalid upstream version: %s in %s",
                        v_str, a.string)
            continue
        v_max = max(v, v_max)
        found = True
    if found:
        return v_max
    return None


def version_from_pypi(name):
    """
    get latest version from PyPI by name
    """
    log.verbose("getting upstream version from PyPI: %s", name)
    url = 'https://pypi.org/pypi/%s/json' % name
    r = requests.get(url)
    if not r.ok:
        return None
    data = r.json()

    versions = data['releases'].keys()
    ver = sorted(versions, key=version.parse)[-1]

    return ver


def latest_apkg_version():
    return version_from_pypi('apkg')



"""
shared apkg testing functions
"""
import os
from pathlib import Path
import shutil


def init_testing_repo(repo_path, test_path):
    dst = Path(test_path) / Path(repo_path).name
    inject_tree(repo_path, dst)
    return dst


def inject_tree(src_path, dst_path, ignore_top_dirs=None):
    """
    copy all files from src_path into dst_path

    overwrite existing files including symlinks
    """
    top_ignore = bool(ignore_top_dirs)
    if not dst_path.exists():
        dst_path.mkdir(parents=True, exist_ok=True)

    # recursively copy all files
    for d, subdirs, files in os.walk(src_path, topdown=True):
        rel_dir = Path(d).relative_to(src_path)
        dst_dir = dst_path / rel_dir
        dst_dir.mkdir(parents=True, exist_ok=True)

        for fn in files:
            src = Path(d) / fn
            dst = dst_dir / fn
            shutil.copy(src, dst)

        if top_ignore:
            # ignore supplied top dirs
            subdirs[:] = [d for d in subdirs if d not in ignore_top_dirs]
            top_ignore = False

        for sd in subdirs:
            # copy symlinks too
            src = Path(d) / sd
            src_str = str(src)
            if os.path.islink(src_str):
                dst = dst_dir / sd
                if dst.exists():
                    dst.unlink()
                linkto = os.readlink(src_str)
                os.symlink(linkto, dst)


def log_contains(string, caplog):
    for r in caplog.records:
        if string in r.message:
            return True
    return False



"""
apkg exceptions

when apkg CLI run results in exception being raised
its returncode is returned as process return code
"""


class ApkgException(Exception):
    msg_fmt = "An unknown error occurred"
    returncode = 1

    def __init__(self, msg=None, returncode=None, **kwargs):
        self.kwargs = kwargs
        if not msg:
            try:
                msg = self.msg_fmt.format(**kwargs)
            except Exception:
                # kwargs doesn't match those in message.
                # Returning this is still better than nothing.
                msg = self.msg_fmt
        if returncode is not None:
            self.returncode = returncode
        super().__init__(msg)


class QuietExit(ApkgException):
    msg_fmt = ""


# 10-29: Invalid
class InvalidUsage(ApkgException):
    msg_fmt = "Invalid usage: {fail}"
    returncode = 10


class InvalidInput(ApkgException):
    msg_fmt = "Invalid input: {fail}"
    returncode = 11


class InvalidType(ApkgException):
    msg_fmt = "Invalid type: $(var) must be {desired} but is {type}"
    returncode = 12


class InvalidChoice(ApkgException):
    msg_fmt = "Invalid choice: {var} must be one of: {opts} (is: {val})"
    returncode = 13


class InvalidVersion(ApkgException):
    msg_fmt = "Invalid version: {ver}"
    returncode = 14


class InvalidFormat(ApkgException):
    msg_fmt = "Invalid format: {fmt}"
    returncode = 16


class InvalidArchiveFormat(InvalidFormat):
    msg_fmt = "Invalid archive format: {fmt}"
    returncode = 17


class InvalidSourcePackageFormat(InvalidFormat):
    msg_fmt = "Invalid source package format: {fmt}"
    returncode = 18


class InvalidCompatLevel(ApkgException):
    msg_fmt = (
        "Project compat level {proj} is newer than current"
        " apkg compat level {apkg}.\n\n"
        "Run `apkg compat` for more information.\n\n"
        "Please upgrade apkg to work with this project."
    )
    returncode = 20


class InvalidVariablesSource(ApkgException):
    msg_fmt = "Invalid variables source: {src}"
    returncode = 26


# 30-39: Missing and NotFound
class MissingRequiredArgument(ApkgException):
    msg_fmt = "Missing required argument: {arg}"
    returncode = 30


class MissingRequiredConfigOption(ApkgException):
    msg_fmt = "Missing required config option: {opt}"
    returncode = 31


class MissingPackagingTemplate(ApkgException):
    msg_fmt = "Missing package template: {temp}"
    returncode = 32


class MissingRequiredModule(ApkgException):
    msg_fmt = "Missing required module: {mod}"
    returncode = 34


class ArchiveNotFound(ApkgException):
    msg_fmt = "Archive not found: {ar}"
    returncode = 36


class SourcePackageNotFound(ApkgException):
    msg_fmt = "{type} source package not found: {srcpkg}"
    returncode = 37


# 40-49: Parsing
class ParsingFailed(ApkgException):
    msg_fmt = "Unable to parse: {fail}"
    returncode = 42

class LintingFailed(ApkgException):
    msg_fmt = "Linting failed: {fail}"
    returncode = 48


# 50-59: remote failures
class FileDownloadFailed(ApkgException):
    msg_fmt = "Failed to download file with code {code}:\n\n{url}"
    returncode = 52


# 60-69: Command
class CommandNotFound(ApkgException):
    msg_fmt = "Command not found: {cmd}"
    returncode = 60


class CommandFailed(ApkgException):
    msg_fmt = (
        "Command failed: {cmdout.args_str} "
        "(return code: {cmdout.returncode})"
    )
    returncode = 62


class UnexpectedCommandOutput(ApkgException):
    msg_fmt = "Unexpected command output: {out}"
    returncode = 64


# 70-79: Distro
class DistroNotSupported(ApkgException):
    msg_fmt = "Distro not supported: {distro}"
    returncode = 70


# 80-89: auto-magic
class UnableToDetectUpstreamVersion(ApkgException):
    msg_fmt = (
        "Unable to detect upstream version.\n\n"
        "Please consider one of following:\n\n"
        "1) set upstream.archive_url\n"
        "2) set upstream.version_script to custom script\n"
        "3) manually supply version using -v/--version option")
    returncode = 84


# 100-110: package tests
class PkgTestFail(ApkgException):
    msg_fmt = "Packaging tests failed."
    returncode = 100



"""
This is convenience interface to apkg.pkgstyles.* modules which can be
dynamically added because apkg.pkgstyles is PEP 420 namespace package.
"""
import importlib
import pkgutil

from apkg import ex
from apkg.log import getLogger
import apkg.pkgstyles


log = getLogger(__name__)


def iter_pkgstyles():
    return pkgutil.iter_modules(
        apkg.pkgstyles.__path__,
        apkg.pkgstyles.__name__ + ".")


def import_pkgstyles():
    styles = {}
    for _, modname, _ in iter_pkgstyles():
        _, _, name = modname.rpartition('.')
        module = importlib.import_module(modname)
        # insert package style name into module for convenience
        module.name = name
        styles[name] = module
    return styles


def get_pkgstyle_for_template(path):
    for style in PKGSTYLES.values():
        if style.is_valid_template(path):
            return style
    return None


def get_pkgstyle_for_distro(distro):
    for style in PKGSTYLES.values():
        for sup_distro in style.SUPPORTED_DISTROS:
            if distro.match(sup_distro):
                return style
    return None


def get_pkgstyle(style):
    return PKGSTYLES.get(style)


def ensure_pkgstyle_fun(pkgstyle, fun):
    f = getattr(pkgstyle, fun, None)
    if not f:
        msg = "%s pkgstyle is missing required function: %s"
    elif not callable(f):
        msg = "%s pkgstyle error: not a function: %s"
    else:
        return f
    raise ex.DistroNotSupported(
        msg % (pkgstyle.__name__, fun))


def call_pkgstyle_fun(pkgstyle, fun, *args, **kwargs):
    f = ensure_pkgstyle_fun(pkgstyle, fun)
    log.verbose("calling %s function: %s", pkgstyle.__name__, fun)
    return f(*args, **kwargs)


PKGSTYLES = import_pkgstyles()



"""
wrapper to select and provide access to available TOML libraries
"""
from apkg import ex


LOAD_LIB = None
DUMP_LIB = None

LOAD_MODE = 'rb'


# get load() and loads()
try:
    from tomllib import load, loads
    LOAD_LIB = 'tomllib'
except ModuleNotFoundError:
    pass

if not LOAD_LIB:
    try:
        from tomli import load, loads
        LOAD_LIB = 'tomli'
    except ModuleNotFoundError:
        pass

if not LOAD_LIB:
    try:
        from toml import load, loads
        LOAD_LIB = 'toml'
        LOAD_MODE = 'r'
    except ModuleNotFoundError:
        pass

if not LOAD_LIB:
    try:
        from tomlkit import load, loads
        LOAD_LIB = 'tomlkit'
    except ModuleNotFoundError:
        pass


# get dump() and dumps()
try:
    from tomli_w import dump, dumps
    DUMP_LIB = 'tomli_w'
except ModuleNotFoundError:
    pass

if not DUMP_LIB:
    try:
        from tomlkit import dump, dumps
        DUMP_LIB = 'tomlkit'
    except ModuleNotFoundError:
        pass

if not DUMP_LIB:
    try:
        from toml import dump, dumps
        DUMP_LIB = 'toml'
    except ModuleNotFoundError:
        pass


def missing_toml_load_module(*args, **kwargs):
    msg=("Requested operation requires TOML load module but none was found.\n\n"
    "Please install one of following Python modules:\n\n"
    "- tomli\n- toml\n- tomlkit")
    raise ex.MissingRequiredModule(msg=msg)


def missing_toml_dump_module(*args, **kwargs):
    msg=("Requested operation requires TOML dump module but none was found.\n\n"
    "Please install one of following Python modules:\n\n"
    "- tomli_w\n- tomlkit\n- toml")
    raise ex.MissingRequiredModule(msg=msg)


# only fail when required functions are called
if not LOAD_LIB:
    load = missing_toml_load_module
    loads = missing_toml_load_module

if not DUMP_LIB:
    dump = missing_toml_dump_module
    dumps = missing_toml_dump_module


def loadp(path):
    """
    Load data from path to TOML file
    """
    with open(path, LOAD_MODE) as f:
        return load(f)



"""
Custom template variable module for apkg adding an enhanced distro variable.

Provides custom template variable `distro_like` that behaves exactly like
`distro`, except that a `distro_like.match()` also matches against the ID_LIKE
information from the distro (e.g. CentOS has rhel in there, ...) respecting the
version as well. As such, version matching should be done only when the derived
distro's version scheme is compatible (e.g. Ubuntu doesn't match Debian's).

This behaviour does not trigger in all cases, if `-d <distro maybe-version>` is
passed to apkg and the distro specified that way is different from the one
inferred from the current host, this new variable will *not* be exposed. This
gives the packager a chance to detect and handle this case however they see
fit, e.g. by setting `distro_like` themselves while preventing running a build
with an untested configuration accidentally.

Enable by adding [[template.variables]] entry to distro/config/apkg.toml:

    [[template.variables]]
    python_module = "apkg.templatevars.distro_like"

Example usage in rpm spec:

{% if distro_like.match('rhel > 9') %}
{# this also matches on Alma 9, CentOS 10, Rocky 28, ... #}
{% endif %}
"""  # noqa

import distro as distro_mod

from apkg import adistro


# called by apkg to get custom variables
def get_variables(env):
    distro = env['distro']
    # poor man's .copy()
    distro_like = adistro.Distro(distro.orig, distro.aliases)

    if distro_mod.id().lower() == distro.id.lower():
        like = set(distro.names + distro_mod.like().split())
        distro_like.names = list(like)
        return {'distro_like': distro_like}

    # Otherwise building for a foreign distro, expose nothing
    return {}



"""
apkg archive (tarball) utils
"""
from pathlib import Path
import shutil
import tempfile

from apkg.parse import split_archive_fn, parse_version

from apkg import ex
from apkg.log import getLogger


log = getLogger(__name__)


def unpack_archive(archive_path, out_path):
    """
    unpack supplied archive into out_path dir

    archive is expected to contain a single root dir

    return path to extracted root dir
    """
    out_path = Path(out_path)
    out_path.mkdir(parents=True, exist_ok=True)
    temp_path = Path(tempfile.mkdtemp(
        prefix='temp_', dir=str(out_path)))
    root_files_old = set(temp_path.glob("*"))
    # shutil doesn't provide a way to check extracted files :(
    # pautil has bugs and got last release in 2016...
    # NOTE(py36): str conversions can be dropped with Python 3.6 support
    shutil.unpack_archive(str(archive_path), str(temp_path))
    root_files_new = set(temp_path.glob("*"))
    root_files = root_files_new - root_files_old
    n_root_files = len(root_files)
    if n_root_files != 1:
        fmt = "Expected a single root dir but instead got %d files in root"
        raise ex.InvalidArchiveFormat(fmt=fmt % n_root_files)
    temp_root = root_files.pop()
    out_root = out_path / temp_root.name
    if out_root.exists():
        log.info("removing existing unpacked archive dir: %s", out_root)
        shutil.rmtree(out_root)
    temp_root.rename(out_root)
    return out_root


def get_archive_version(archive_path):
    """
    return archive version detected from archive name
    """
    archive_path = Path(archive_path)
    _, _, ver, _ = split_archive_fn(archive_path.name)
    return parse_version(ver)



import click
from packaging import version

from apkg import __version__
from apkg import compat as compat_
from apkg import ex
from apkg.log import getLogger, T
from apkg.project import Project
from apkg.util.upstreamversion import latest_apkg_version


log = getLogger(__name__)


@click.command(name='compat')
@click.help_option('-h', '--help', help='show this help')
@click.option('--latest/--no-latest', default=True, show_default=True,
              help="check latest apkg version/compat level online")
@click.option('-n', '--notes', default=0,
              help="only show compat level upgrade notes since level")
def cli_compat(*args, **kwargs):
    """
    check apkg compat level

    display hints on howto set it properly if it isn't current

    return exit code 1 if compat check failed
    """
    ok = compat(*args, **kwargs)
    if not ok:
        raise ex.QuietExit()


def compat(latest=True, notes=None):
    current_compat = compat_.COMPAT_LEVEL

    if notes:
        notes_ = compat_.get_level_notes(notes, current_compat)
        print_notes(notes_)
        return True

    proj = Project(auto_compat=False)
    proj_compat = proj.compat_level
    current_version = version.parse(__version__)

    if latest:
        latest_version = version.parse(latest_apkg_version())
        latest_compat = compat_.get_upstream_compat_level(latest_version) or 1

    msg = "project compat level:       {t.bold}{level}{t.normal}"
    print(msg.format(level=proj_compat or 'N/A', t=T))

    msg = "current apkg compat level:  {t.bold}{level}{t.normal}"
    print(msg.format(level=current_compat, t=T))

    if latest:
        msg = "latest apkg compat level:   {t.bold}{level}{t.normal}"
        print(msg.format(level=latest_compat, t=T))

    print()

    msg = "current apkg version:  {t.bold}{v}{t.normal}"
    print(msg.format(v=current_version, t=T))

    if latest:
        msg = "latest apkg version:   {t.bold}{v}{t.normal}"
        print(msg.format(v=latest_version, t=T))

    print()

    if latest and current_compat < latest_compat:
        print("latest apkg-{v} has newer compat level"
              " {l} - consider upgrading apkg\n".format(
                  l=latest_compat, v=latest_version))

    if not proj_compat:
        print("{t.bold_yellow}⚠ apkg.compat level isn't set in config"
              "{t.normal}".format(t=T))
        print("\nPlease consider adding\n\n"
              "    {t.command}[apkg]\n"
              "    compat = {ccl}{t.normal}\n\n"
              "to project config: {t.bold}{cfg}{t.normal}\n".format(
                  ccl=current_compat, cfg=proj.path.config, t=T))
        print("This will ensure compatibility between apkg versions.")
        return False

    if proj_compat < current_compat:
        print("{t.bold_yellow}⚠ project compat level {pcl} is older than"
              " current {ccl}{t.normal}".format(
                  pcl=proj_compat, ccl=current_compat, t=T))
        print("\nPlease consider bumping\n\n"
              "    {t.command}[apkg]\n"
              "    compat = {ccl}{t.normal}\n\n"
              "in project config: {t.bold}{cfg}{t.normal}".format(
                  ccl=current_compat, cfg=proj.path.config, t=T))
        print("\nInspect following upgrade notes:")
        print_notes(compat_.get_level_notes(proj_compat, current_compat))
        return False

    if proj_compat > current_compat:
        print("{t.bold_red}❌ project compat level {pl} is newer than"
              " current {cl} - please upgrade apkg!{t.normal}".format(
                  pl=proj_compat, cl=current_compat, t=T))
        return False

    print("{t.bold_green}✔ project compat level {l} matches installed"
          " apkg version{t.normal}".format(l=current_compat, t=T))
    return True


def print_notes(notes):
    for level, (ver, note) in notes.items():
        print("\n{t.magenta}# COMPAT LEVEL {l}\n\n"
              "{t.normal}Introduced in apkg-{v}\n".format(v=ver, l=level, t=T))
        print(note)


APKG_CLI_COMMANDS = [cli_compat]



"""
module for running apkg packaging tests
"""
import email
import glob
import os
import re
import shutil

from pathlib import Path
try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property

import jinja2

from apkg import adistro
from apkg import ex
from apkg.log import getLogger
from apkg import pkgstyle as _pkgstyle
from apkg.pkgtemplate import IncludeRawExtension
from apkg.util.run import run, sudo
from apkg.util.test import inject_tree


log = getLogger(__name__)


# tests extra selection types ordered by priority
TES_ALIAS, TES_DISTRO, TES_PKGSTYLE, TES_ALL = range(4)
TESTS_EXTRA_SELECTION_STR = {
    TES_ALIAS: 'distro alias',
    TES_DISTRO: 'distro-specific',
    TES_PKGSTYLE: 'pkgstyle default',
    TES_ALL: 'default fallback',
}


class TestConfig:
    def __init__(self,
                 name=None,
                 test_command=None,
                 tests_dir=None,
                 deps=None,
                 restrictions=None):
        self.name = name
        self.test_command = test_command
        self.tests_dir = tests_dir
        self.deps = deps if deps else []
        self.restrictions = restrictions if restrictions else []

    def __str__(self):
        if self.test_command:
            return '$ %s' % self.test_command
        return self.name


def parse_test_config(test_text):
    tests = []
    deps = []
    restrictions = []

    msg = email.message_from_string(test_text)
    tests_text = msg['Tests']
    test_command_text = msg['Test-Command']
    tests_dir = msg['Tests-Directory']
    if tests_dir:
        tests_dir = Path(tests_dir)
    deps_text = msg['Depends']
    restrictions_text = msg['Restrictions']

    if deps_text:
        deps = re.split(r'[\s\n]*,[\s\n]*', deps_text.strip())
        # @ means "all packages built from source package"
        # this is implicit with apkg tests
        deps = [d for d in deps if d and d != '@']

    if restrictions_text:
        restrictions = re.split(r'[\s\n]+', restrictions_text.strip())

    if not (tests_text or test_command_text):
        log.warning("Test missing both Tests and Test-Command fields"
                    " - skipping. Test text:\n%s", test_text)
        return tests

    if test_command_text:
        tc = TestConfig(
            test_command=test_command_text,
            tests_dir=tests_dir,
            deps=deps,
            restrictions=restrictions,
        )
        tests.append(tc)

    if tests_text:
        for tn in re.split(r'[\s\n,]+', tests_text.strip()):
            tc = TestConfig(
                name=tn,
                tests_dir=tests_dir,
                deps=deps,
                restrictions=restrictions,
            )
            tests.append(tc)

    return tests


class TestsExtra:
    def __init__(self, path,
                 selection=TES_DISTRO,
                 distro_rules=None):
        self.path = path
        self.selection = selection
        self.distro_rules = distro_rules

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, val):
        self._path = Path(val)
        self._name = self._path.name

    @property
    def name(self):
        return self._name

    @cached_property
    def control_path(self):
        return self._path / 'control'

    @cached_property
    def has_extra_files(self):
        for f in self._path.iterdir():
            if f.name == 'control':
                continue
            return True
        return False

    def selection_str(self):
        return TESTS_EXTRA_SELECTION_STR.get(self.selection, 'INVALID')

    def match_distro(self, distro):
        return self.distro_rules.match(distro)

    @cached_property
    def env(self):
        return jinja2.Environment(
            loader=jinja2.FileSystemLoader('.'),
            extensions=[IncludeRawExtension])

    def render(self, out_path, tvars):
        """
        render tests extra files into specified output directory

        Args:
            out_path: output base path
            tvars: variables available from extra templates
        """
        log.info("renderding tests extra: %s -> %s", self.path, out_path)
        if out_path.exists():
            log.verbose("tests extra output dir exists: %s", out_path)
        else:
            out_path.mkdir(parents=True, exist_ok=True)

        # recursively render all files
        for d, _, files in os.walk(self.path, followlinks=True):
            rel_dir = Path(d).relative_to(self.path)
            dst_dir = out_path / rel_dir
            dst_dir.mkdir(parents=True, exist_ok=True)

            for fn in files:
                dst = out_path / rel_dir / fn
                src = Path(d) / fn

                log.verbose("rendering tests extra file: %s -> %s", src, dst)
                t = self.env.get_template(str(src))
                with dst.open('w') as dstf:
                    dstf.write(t.render(**tvars))

                # preserve original permission
                dst.chmod(src.stat().st_mode)

    def render_file_content(self, name, tvars):
        """
        render tests extra file in memory and return its content
        """
        src = self.path / name
        t = self.env.get_template(str(src))
        return t.render(**tvars)

    def __repr__(self):
        return "TestsExtra<%s>" % self.name


class PackageTests:
    def __init__(self, tests_in_path, tests_out_path, extra, distro):
        self.tests_in_path = tests_in_path
        self.tests_out_path = tests_out_path
        self.tests_path = tests_in_path
        self.control_path = tests_in_path / 'control'
        self.distro = distro
        self.extra = extra
        self.extra_control = False
        if self.extra and self.extra.control_path.exists():
            self.control_path = self.extra.control_path
            self.extra_control = True

    @cached_property
    def tvars(self):
        """
        vars availabe during templating
        """
        return {
            'distro': self.distro,
        }

    @property
    def extra_path(self):
        if not self.extra:
            return None
        return self.extra.path

    @cached_property
    def control_text(self):
        if self.extra_control:
            return self.extra.render_file_content(
                'control', self.tvars)
        if not self.control_path.exists():
            return ''
        return self.control_path.open('rt').read().strip()

    @cached_property
    def tests(self):
        tests = []
        if not self.control_text:
            return []
        for txt in re.split(r'\n\s*\n[\s\n]*', self.control_text.strip()):
            tests.extend(parse_test_config(txt))
        return tests

    @cached_property
    def deps(self):
        d = set()
        for tc in self.tests:
            for dep in tc.deps:
                d.add(dep)
        return sorted(d)

    def prepare(self):
        """
        prepare packaging tests for run
        """
        if self.extra and self.extra.has_extra_files:
            # extra with overlay files - need to copy over
            self.tests_path = self.tests_out_path / self.distro.idver / 'tests'
            log.info("preparing tests instance with extra files: %s -> %s",
                     self.tests_in_path, self.tests_path)
            if self.tests_path.exists():
                log.verbose("removing existing tests instance: %s",
                            self.tests_path)
                shutil.rmtree(self.tests_path)
            else:
                self.tests_path.parent.mkdir(exist_ok=True, parents=True)
            inject_tree(self.tests_in_path, self.tests_path,
                        ignore_top_dirs=['extra'])
            self.extra.render(self.tests_path, self.tvars)
            return

        # inline tests (no overlay files from extra)
        log.info("using inline tests: %s", self.tests_path)
        self.tests_path = self.tests_in_path

    def filter_tests(self, test_filter):
        if not test_filter:
            return self.tests
        return [t for t in self.tests if re.search(test_filter, str(t))]

    def run(self, test_filter=None):
        """
        run packaging tests
        """
        self.prepare()
        n_all_tests = len(self.tests)
        if n_all_tests < 1:
            msg = "No tests found in: %s" % self.tests_path
            raise ex.InvalidUsage(msg=msg)
        tests = self.filter_tests(test_filter)
        if test_filter:
            log.info("filtering tests by regex: %s", test_filter)
            if not tests:
                msg = "No tests match regex filter: %s" % test_filter
                raise ex.InvalidUsage(msg=msg)
        n_tests = len(tests)
        log.info("running %s of %s tests...",
                 n_tests, n_all_tests)

        tests_ok = []
        tests_fail = []
        tests_skip = []
        for tc in tests:
            shell = False
            if tc.name:
                tests_path = tc.tests_dir or self.tests_path
                cmd = str(tests_path / tc.name)
            elif tc.test_command:
                cmd = tc.test_command
                shell = True
            else:
                log.warning("skipping invalid test")
                continue

            log.bold("test RUN: %s", cmd)
            try:
                if 'needs-root' in tc.restrictions:
                    _run = sudo
                else:
                    _run = run
                out = _run(cmd, shell=shell, check=False,
                           log_fun=log.verbose_command)
                if out.returncode == 77:
                    log.warning("test SKIP: %s (exit code: %s)",
                                cmd, out.returncode)
                    tests_skip.append(tc)
                elif out.returncode != 0:
                    if 'flaky' in tc.restrictions:
                        log.warning("test SKIP FLAKY: %s (exit code: %s)",
                                    cmd, out.returncode)
                        tests_skip.append(tc)
                    else:
                        log.error("test FAIL: %s (exit code: %s)",
                                  cmd, out.returncode)
                        tests_fail.append(tc)
                else:
                    log.success("test OK: %s", cmd)
                    tests_ok.append(tc)
            except ex.CommandNotFound:
                log.error("test FAIL: %s (command not found)", cmd)
                tests_fail.append(tc)

        if tests_fail:
            log.error("%s tests failed", len(tests_fail))
            return False
        if not tests_ok:
            log.error("no tests were run")
            return False
        if tests_skip:
            log.success("%s tests succeeded, %s skipped",
                        len(tests_ok), len(tests_skip))
        else:
            log.success("%s tests succeeded", len(tests_ok))

        return True


def load_tests_extras(path, distro_aliases=None):
    """
    load package tests' extras sorted by evaluation priority

    Params:
        path - tests extras base path (i.e.: distro/tests/extra)
        distro_aliases - distro aliases dict (optional)

    Returns:
        list of TestsExtras in order of evaluation
    """
    aliases = distro_aliases or {}

    alias_exs = []
    distro_exs = []
    pkgstyle_exs = []
    default_ex = None

    for entry_path in glob.glob('%s/*' % path):
        if not os.path.isdir(entry_path):
            # ignore non-dirs
            continue

        extra = TestsExtra(entry_path)

        alias_rules = aliases.get(extra.name)
        if alias_rules:
            # distro alias (name match)
            extra.selection = TES_ALIAS
            extra.distro_rules = alias_rules
            alias_exs.append(extra)
            continue

        dstyle = _pkgstyle.PKGSTYLES.get(extra.name)
        if dstyle:
            # pkgstyle default (name match)
            extra.selection = TES_PKGSTYLE
            extra.distro_rules = adistro.distro_rules(dstyle.SUPPORTED_DISTROS)
            pkgstyle_exs.append(extra)
            continue

        if extra.name == 'all':
            # default fallback (special 'all' name)
            extra.selection = TES_ALL
            extra.distro_rules = adistro.DistroRuleAll()
            default_ex = extra
            continue

        # distro-specific
        extra.distro_rules = adistro.name2rule(extra.name)
        distro_exs.append(extra)

    distro_exs = adistro.sort_by_name(distro_exs)
    alias_exs.sort(key=lambda x: x.name)
    pkgstyle_exs.sort(key=lambda x: x.name)
    extras = alias_exs + distro_exs + pkgstyle_exs
    if default_ex:
        extras.append(default_ex)

    return extras



"""
apkg logging module

introduces custom log levels and associated formatting using standard python
logging module

different formatting for individual log levels including optional color support

logs to stderr by default and uses colors if blessing are available and
terminal supports it, use set_color() to control

to get this nice colorful logging to terminal stderr anywhere in python:

```
from apkg.log import getLogger

log.getLogger(__name__)

log.info("log message")
log.success("great %s", "success")
```
"""
import logging
# import getLogger for convenience so that one can use
#
#   from apkg.log import getLogger
#
#   log = getLogger(__name__)
#
# to init apkg logging without the need to import logging module
# while making sure logging.setLoggerClass(ApkgLogger) was called
# (see below in this module) so that extra log level functions
# are available
from logging import getLogger  # noqa

import sys
from apkg import terminal


# add custom log levels and handle formatting
# including color based on log level
ERROR = logging.ERROR
WARN = logging.WARN
INFO = logging.INFO
VERBOSE = int((logging.INFO + logging.DEBUG) / 2)
DEBUG = logging.DEBUG
SUCCESS = WARN + 1
SUDO = WARN - 1
BOLD = INFO + 2
COMMAND = INFO + 1
VERBOSE_COMMAND = VERBOSE + 1


LOG_LEVEL_FORMATS_META = {
    ERROR: ('E', 'error'),
    SUCCESS: ('✓', 'success'),
    WARN: ('W', 'warn'),
    SUDO: ('#', 'sudo'),
    BOLD: ('I', 'bold'),
    INFO: ('I', None),
    VERBOSE: ('V', None),
    DEBUG: ('D', None),
    COMMAND: ('$', 'command'),
    VERBOSE_COMMAND: ('$', 'command'),
}


# current global log level - DON'T IMPORT/USE THIS DIRECTLY
# get LOG_LEVEL using get_log_level()
# set LOG_LEVEL using set_log_level()
LOG_LEVEL = INFO

# stacklevel arg introduced in Python 3.8 is required to display proper
# function name due to helpers being used
# TODO: drop this when Python 3.7 support ends
HAS_STACKLEVEL = bool(sys.version_info >= (3, 8))


def get_log_level():
    """
    get global log level (int)
    """
    return LOG_LEVEL


class ApkgLogger(logging.Logger):
    """
    standard logger with extra functions for custom log levels
    """
    def success(self, msg, *args, **kwargs):
        if self.isEnabledFor(SUCCESS):
            if HAS_STACKLEVEL:
                kwargs['stacklevel'] = 2
            self._log(SUCCESS, msg, args, **kwargs)

    def bold(self, msg, *args, **kwargs):
        if self.isEnabledFor(BOLD):
            if HAS_STACKLEVEL:
                kwargs['stacklevel'] = 2
            self._log(BOLD, msg, args, **kwargs)

    def verbose(self, msg, *args, **kwargs):
        if self.isEnabledFor(VERBOSE):
            if HAS_STACKLEVEL:
                kwargs['stacklevel'] = 2
            self._log(VERBOSE, msg, args, **kwargs)

    def command(self, msg, *args, **kwargs):
        if self.isEnabledFor(COMMAND):
            if HAS_STACKLEVEL:
                kwargs['stacklevel'] = 3
            self._log(COMMAND, msg, args, **kwargs)

    def verbose_command(self, msg, *args, **kwargs):
        if self.isEnabledFor(VERBOSE_COMMAND):
            if HAS_STACKLEVEL:
                kwargs['stacklevel'] = 3
            self._log(VERBOSE_COMMAND, msg, args, **kwargs)

    def sudo(self, msg, *args, **kwargs):
        if self.isEnabledFor(SUDO):
            if HAS_STACKLEVEL:
                kwargs['stacklevel'] = 4
            self._log(SUDO, msg, args, **kwargs)


class LogTerminal(terminal.Terminal):
    """
    helper for handling colors when logging to terminal
    """
    @property
    def warn(self):
        return self.yellow

    @property
    def important(self):
        return self.yellow_bold

    @property
    def error(self):
        return self.red

    @property
    def success(self):
        return self.green

    @property
    def command(self):
        return self.cyan

    @property
    def sudo(self):
        return self.magenta


# global terminal instance for color printing
T = LogTerminal()


def set_colors(colors):
    """
    enable/disable colors in apkg logging terminal
    """
    global T
    if colors == 'yes':
        if not terminal.COLOR_TERMINAL:
            return False
        T = LogTerminal(force_styling=True)
        return True
    elif colors == 'no':
        if not terminal.COLOR_TERMINAL:
            return True
        T = LogTerminal(force_styling=None)
        return True
    elif colors == 'auto':
        T = LogTerminal()
        return True
    return False


def logfmt(prefix, color, levelno):
    """
    helper to generate loggin.Formatter compatible log format
    based on current log level
    """
    if color:
        color_fun = getattr(T, color)
    else:
        def color_fun(x):
            return x
    if levelno <= DEBUG:
        # prepended extra line: path, line number and function
        fmt = '%s:%s @ %s:\n%s' % (
            T.magenta('%(pathname)s'),
            T.yellow('%(lineno)s'),
            T.command('%(funcName)s()'),
            color_fun('%s %%(message)s' % prefix))
    elif levelno <= VERBOSE:
        # insert module and function name
        fmt = '%s %s.%s: %s' % (
            color_fun(prefix),
            T.magenta('%(module)s'),
            T.command('%(funcName)s()'),
            color_fun('%(message)s'))
    else:
        fmt = '%s %%(message)s' % prefix
        fmt = color_fun(fmt)
    return fmt


def render_log_level_formats(levelno):
    fmts = {}
    for level, entry in LOG_LEVEL_FORMATS_META.items():
        prefix, color = entry
        fmts[level] = logfmt(prefix, color, levelno)
    return fmts


# initialize global log formats
LOG_LEVEL_FORMATS = render_log_level_formats(LOG_LEVEL)


class ApkgLogFormatter(logging.Formatter):
    """
    default apkg log formatter with support for custom formats
    per log level including custom colors
    """
    def format(self, record):
        # change format depending on log level
        fmt = LOG_LEVEL_FORMATS.get(record.levelno, "? %(message)s")
        # this hacky but mostly reliable
        # pylint: disable=protected-access
        self._style._fmt = fmt
        # use default formatter with custom format
        return super().format(record)


# logging setup
logging.setLoggerClass(ApkgLogger)
handler = logging.StreamHandler(sys.stderr)
handler.setFormatter(ApkgLogFormatter())
logging.root.addHandler(handler)
logging.root.setLevel(logging.INFO)


def set_log_level(levelno):
    """
    set global log level and regenerate log formats accordingly
    """
    global LOG_LEVEL
    global LOG_LEVEL_FORMATS

    LOG_LEVEL = levelno
    logging.root.setLevel(LOG_LEVEL)
    LOG_LEVEL_FORMATS = render_log_level_formats(LOG_LEVEL)



import re
import requests

from apkg import ex
from apkg.log import getLogger
from apkg.util.upstreamversion import latest_apkg_version


log = getLogger(__name__)


# current apkg compatibility level
COMPAT_LEVEL = 6


COMPAT_LEVEL_NOTES = {
    6: ('0.7.0', """Forward incompatible update, changes required:

## new make_archive_script and version_script YAML interface

make_archive_script and version_script stdout is now expected to be
in YAML format, any messages should go to stderr.

Edit your make_archive_script to output like this:

archive: pkg/archives/dev/banana-1.2.3.tar.gz

Edit your upstream.version_script (if any) to output like this:

version: 1.2.3"""),
    5: ('0.6.0', """Forward compatible update for most users.

## RPM macros in .spec templates are now evaluated using rpmspec if available

This might result in different build deps being detected by apkg when using
%if macros on BuildRequires. Results depend on rpmspec availability
and macros defined on the host machine (such as %fedora)."""),
    4: ('0.5.0', """Forward compatible update, no action required."""),
    3: ('0.4.0', """Forward compatible update, no action required."""),
    2: ('0.3.0', """Forward compatible update for most users.

## new flexible template selection mechanism

Make sure each template dir in distro/pkg/ matches name
of a respective pkgstyle such as deb, rpm, arch, or nix.

Inspect new improved `apkg status` package templates listing.""")
}


def ensure_compat(compat_level):
    if not compat_level:
        return
    compat_level = int(compat_level)
    if COMPAT_LEVEL < compat_level:
        raise ex.InvalidCompatLevel(
            proj=compat_level, apkg=COMPAT_LEVEL)


def level_status(compat_level):
    if compat_level is None:
        return False, 'not set'
    elif compat_level < COMPAT_LEVEL:
        return False, 'old'
    elif compat_level > COMPAT_LEVEL:
        return False, 'new'
    elif compat_level == COMPAT_LEVEL:
        return True, 'current'
    else:
        return False, 'error'


def get_level_notes(old_level, new_level=COMPAT_LEVEL):
    """
    return a dict of relevant compat level notes
    for migration from old_level to new_level.
    """
    actions = {}
    for i in range(old_level + 1, new_level + 1):
        actions[i] = COMPAT_LEVEL_NOTES[i]
    return actions


def get_upstream_compat_level(apkg_version=None):
    """
    fetch upstream compat level for apkg version

    use latest version when apkg_version isn't specified
    """
    if not apkg_version:
        apkg_version = latest_apkg_version()

    url = 'https://gitlab.nic.cz/packaging/apkg/-/raw/v%s/apkg/compat.py'
    url = url % apkg_version
    log.verbose("getting upstream compat level for apkg-%s: %s",
                apkg_version, url)

    r = requests.get(url)
    if not r.ok:
        return None
    m = re.search(r'^COMPAT_LEVEL\s*=\s*(\d+)\s*$', r.text, re.MULTILINE)
    if not m:
        return None
    level = int(m.group(1))
    return level



"""
apkg package style for RPM-based distros
such as Fedora, CentOS, SUSE, RHEL.

**source template**: `*.spec` and friends

**source package:** `*.src.rpm`

**packages:** `*.rpm`

**required distro packages**:

 * core: `rpm-build`
 * isolated build: `mock`

**template variables:**

 * `now`: current date in RPM changelog format,
          use `SOURCE_DATE_EPOCH` env var when set
"""
import glob
import os
from pathlib import Path
import re
import shutil
import subprocess
import time

from packaging.version import Version

from apkg import ex
from apkg.util import common
from apkg.log import getLogger
from apkg.util.run import run, sudo


log = getLogger(__name__)


EL_FAMILY_DISTROS = [
    "almalinux",
    "centos",
    "oracle",
    "rhel",
    "rocky",
    "scientific",
]
SUPPORTED_DISTROS = sorted(EL_FAMILY_DISTROS + [
    "fedora",
    "opensuse",
    "pidora",
])
DISTRO_REQUIRES = {
    'core': ['rpm-build'],
    'isolated': ['mock'],
    'lint': ['rpmlint'],
}


def get_now():
    """
    current date and time in RPM changelog format

    SOURCE_DATE_EPOCH env variable overrides the time when set
    """
    return time.strftime(
        "%a %b %d %Y",
        time.gmtime(int(os.environ.get('SOURCE_DATE_EPOCH', time.time())))
    )


TEMPLATE_VARS_DYNAMIC = {
    'now': get_now,
}


RE_PKG_NAME = r'Name:\s*(\S+)'
RE_BUILD_REQUIRES = r'BuildRequires:\s*(.*)'
RE_RPMBUILD_OUT_RPM = r'Wrote:\s+(.*\.rpm)\s*'
RE_RPMBUILD_OUT_SRPM = r'Wrote:\s+(.*\.src.rpm)\s*'


def is_valid_template(path):
    return bool(get_spec_(path))


def get_template_name(template, distro=None):
    """
    get Name from .spec
    """
    spec_text = render_spec_from_template_(template, distro=distro)
    for line in spec_text.splitlines():
        m = re.match(RE_PKG_NAME, line)
        if m:
            name = m.group(1)
            if has_rpm_macros_(name):
                raise ex.ParsingFailed(
                    msg="unable to parse RPM macro in Name: %s" % name)
            return name

    raise ex.ParsingFailed(
        msg="unable to determine Name from rpm template: %s" % template.path)


def get_srcpkg_nvr(path):
    name = str(path.name)
    m = re.match(r'(.*)\.src.rpm', name)
    if m:
        return m.group(1)
    return name


def build_srcpkg(
        build_path,
        out_path,
        archive_info,
        template,
        tvars):
    """
    build .src.rpm source package
    """
    rpmbuild_topdir = build_path / 'rpmbuild'
    rpmbuild_src = rpmbuild_topdir / 'SOURCES'
    rpmbuild_spec = rpmbuild_topdir / 'SPEC'

    rpmbuild_src.mkdir(parents=True, exist_ok=True)

    template.render(rpmbuild_src, tvars=tvars)

    spec_src_path = get_spec_(rpmbuild_src)
    spec_path = rpmbuild_spec / spec_src_path.name
    log.verbose("moving .spec file into SPEC: %s", spec_path)
    rpmbuild_spec.mkdir(exist_ok=True)
    spec_src_path.rename(spec_path)

    log.info("copying archive files into SOURCES: %s", rpmbuild_src)
    archive_paths = [archive_info["archive"]]
    archive_paths.extend(archive_info.get("components", {}).values())
    for src_path in archive_paths:
        dst_path = rpmbuild_src / src_path.name
        shutil.copyfile(src_path, dst_path)
    log.info("building .src.rpm using rpmbuild")
    out = run('rpmbuild', '-bs',
              '--define', '_topdir %s' % rpmbuild_topdir.resolve(),
              spec_path)

    log.info("copying .src.rpm to result dir: %s", out_path)
    out_path.mkdir(parents=True)
    srcpkgs = []
    for m in re.finditer(RE_RPMBUILD_OUT_SRPM, out):
        srpm = m.group(1)
        src_srpm = Path(srpm)
        dst_srpm = out_path / src_srpm.name
        shutil.copyfile(src_srpm, dst_srpm)
        srcpkgs.append(dst_srpm)
    if not srcpkgs:
        raise ex.ParsingFailed(
            msg="unable to parse rpmbuild results")
    if len(srcpkgs) > 1:
        raise ex.UnexpectedCommandOutput(
            msg="rpmbuild produced multiple .src.rpm files")
    return srcpkgs


def build_packages(
        build_path,
        out_path,
        srcpkg_paths,
        **kwargs):
    """
    build .rpm packages from .src.rpm
    """
    isolated = kwargs.get('isolated')
    pkgs = []
    srcpkg_path = srcpkg_paths[0]

    if isolated:
        log.info("starting isolated .rpm build using mock")
        # sudo shouldn't be necessary when in mock group
        # but apkg can't rely on that especially in containers etc.
        sudo('mock',
             '--resultdir', build_path,
             srcpkg_path,
             preserve_env=True)
        log.info("copying built packages to result dir: %s", out_path)
        out_path.mkdir(parents=True)
        for rpm in glob.iglob('%s/*.rpm' % build_path):
            src_pkg = Path(rpm)
            dst_pkg = out_path / src_pkg.name
            shutil.copyfile(src_pkg, dst_pkg)
            pkgs.append(dst_pkg)
    else:
        log.info("starting direct host .rpm build using rpmbuild")
        rpmbuild_topdir = build_path / 'rpmbuild'
        rpmbuild_topdir.mkdir(parents=True, exist_ok=True)
        out = run('rpmbuild', '--rebuild',
                  '--define', '_topdir %s' % rpmbuild_topdir.resolve(),
                  srcpkg_path)
        log.info("copying built packages to result dir: %s", out_path)
        out_path.mkdir(parents=True)
        for m in re.finditer(RE_RPMBUILD_OUT_RPM, out):
            rpm = m.group(1)
            src_pkg = Path(rpm)
            dst_pkg = out_path / src_pkg.name
            shutil.copyfile(src_pkg, dst_pkg)
            pkgs.append(dst_pkg)
        if not pkgs:
            raise ex.ParsingFailed(
                msg="unable to parse rpmbuild results")

    return pkgs


def install_build_deps(
        deps,
        **kwargs):
    # dnf/zypper install handles build deps
    install_distro_packages(deps, **kwargs)


def install_custom_packages(
        packages,
        **kwargs):
    # dnf/zypper install handles local packages
    kwargs['allow_unsigned'] = True
    install_distro_packages(packages, **kwargs)


def install_distro_packages(
        packages,
        **kwargs):
    allow_unsigned = kwargs.get('allow_unsigned', False)
    interactive = kwargs.get('interactive', False)
    distro = kwargs.get('distro')
    reinstall = kwargs.get('reinstall', False)
    force = kwargs.get('force', False)
    pm = get_package_manager_(distro)

    cmd = [pm]
    if pm == 'zypper':
        # zypper (openSUSE)
        cmd += ['install']
        if reinstall or force:
            # exclusive with -C
            cmd += ['--force']
        else:
            # use zypper capabilities
            cmd += ['-C']
        if force:
            cmd += ['--replacefiles']
        if allow_unsigned:
            cmd += ['--allow-unsigned-rpm']
    else:
        # dnf (Fedora / EL)
        if reinstall:
            cmd += ['reinstall']
        else:
            cmd += ['install']
        if force:
            cmd += ['--allowerasing']

    if not interactive:
        cmd += ['-y']
    cmd += packages
    sudo(cmd)


def get_build_deps_from_template(
        template,
        distro=None):
    """
    parse BuildRequires from packaging template
    """
    spec_text = render_spec_from_template_(template, distro=distro)
    return get_build_deps_from_spec_(spec_text)


def get_build_deps_from_srcpkg(
        srcpkg_path,
        **_):
    """
    parse BuildRequires from .src.rpm
    """
    cmd = "rpm2cpio '%s' | cpio -i --to-stdout '*.spec'" % srcpkg_path
    spec_text = subprocess.getoutput(cmd)
    return get_build_deps_from_spec_(spec_text)


def lint(
        pkg_paths,
        info=False,
        strict=False,
        distro=None,
        **kwargs):
    """
    lint files using rpmlint
    """
    log.info('linting %s files with rpmlint', len(pkg_paths))
    suse = distro and distro.id == 'opensuse'
    cmd = ['rpmlint']
    if info:
        cmd += ['--info']
    if strict and not suse:
        cmd += ['--strict']
    cmd += pkg_paths
    o = run(cmd, check=False, direct=True)
    return o.returncode


# functions bellow with _ postfix are specific to this pkgstyle


def get_package_manager_(distro):
    if distro:
        if distro.id == 'opensuse':
            # use zypper on openSUSE
            return 'zypper'
        if (distro.id in EL_FAMILY_DISTROS
                and distro.version
                and Version(distro.version) <= Version("7")):
            # use yum on EL <= 7
            return 'yum'
    # use dnf by default
    if not shutil.which('dnf'):
        # dnf not available - check for microdnf (as seen in EL 10 minimal images)
        if shutil.which('microdnf'):
            return 'microdnf'
    return 'dnf'


def get_spec_(path):
    for s in glob.iglob("%s/*.spec" % path):
        return Path(s)
    return None


def render_spec_from_template_(template, distro=None, parse_macros=True):
    """
    render spec file from template
    """
    spec_path = get_spec_(template.path).relative_to(template.path)
    tvars = {}
    if distro:
        tvars['distro'] = distro
    spec_txt = template.render_file_content(spec_path, tvars=tvars)
    if parse_macros:
        spec_txt = parse_spec_(spec_txt)
    return spec_txt


def parse_spec_(spec_text):
    """
    parse spec to expand macros if rpmbuild is available
    """
    # done through temp file because rpmspec doesn't work
    # on /dev/stdin on openSUSE for some reason :-/
    with common.text_tempfile(spec_text, prefix='apkg_rpm.spec_') as spec_path:
        try:
            spec_parsed = run('rpmspec', '-P', spec_path, quiet=True)
        except ex.CommandNotFound:
            log.warning("rpmspec not available - unable to parse RPM macros in .spec")
            log.info("consider running `apkg system-setup` to get rpmspec"
                     " or installing rpm-build package manually")
            return spec_text

    return spec_parsed


def get_build_deps_from_spec_(spec_text):
    """
    parse BuildRequires from .spec file content
    """
    return re.findall(RE_BUILD_REQUIRES, spec_text)


def has_rpm_macros_(s):
    """
    does string contain RPM macros? (patterns like %{name} or %name)
    """
    return bool(re.search(r'%\{?\w+\}?', s))



from pathlib import Path
import shutil

import click

from apkg import adistro
from apkg.cache import file_checksum, path_checksum
from apkg import ex
from apkg.util import common
from apkg.commands.get_archive import get_archive
from apkg.commands.make_archive import make_archive, sanitize_archive_output
from apkg.log import getLogger
from apkg.pkgstyle import call_pkgstyle_fun
from apkg.project import Project
from apkg.util.archive import get_archive_version


log = getLogger(__name__)


@click.command(name="srcpkg")
@click.argument('inputs', nargs=-1)
@click.option('-a', '--archive', is_flag=True,
              help="source package from speficied archive file(s)")
@click.option('-u', '--upstream', is_flag=True,
              help="upstream source package from archive templates")
@click.option('-v', '--version',
              help=("upstream archive version to use"
                    ", implies --upstream, exclusive with --archive"))
@click.option('-r', '--release',
              help="set packagge release  [default: 1]")
@click.option('-d', '--distro',
              help="set target distro  [default: current]")
@click.option('-O', '--result-dir',
              help=("put results into specified dir"
                    "  [default: pkg/srcpkg/DISTRO/NVR]"))
@click.option('--render-template', is_flag=True,
              help="only render source package template")
@click.option('--cache/--no-cache', default=True, show_default=True,
              help="enable/disable cache")
@click.option('-F', '--in-file', 'in_files', multiple=True,
              help="specify input file(s), '-' to read from stdin")
@click.option('-f', '--in-format', default='auto', show_default=True,
              type=click.Choice(['auto', 'yaml', 'list']),
              help="set input format")
@click.help_option('-h', '--help',
                   help="show this help message")
def cli_srcpkg(*args, **kwargs):
    """
    create source package (files to build packages from)
    """
    results = srcpkg(*args, **kwargs)
    common.print_results(results)
    return results


def srcpkg(
        archive=False,
        inputs=None,
        in_files=None,
        in_format=None,
        upstream=False,
        version=None,
        release=None,
        distro=None,
        result_dir=None,
        render_template=False,
        cache=True,
        project=None):
    """
    create source package
    """
    srcpkg_type = 'upstream' if upstream else 'dev'
    if render_template:
        log.bold('rendering %s source package template', srcpkg_type)
    else:
        log.bold('creating %s source package', srcpkg_type)

    if version:
        # --version implies --upstream
        upstream = True
        if archive:
            raise ex.InvalidInput(
                fail="--archive and --version options are mutually exclusive")

    proj = project or Project()
    distro = adistro.distro_arg(distro, proj)
    log.info("target distro: %s", distro)

    if render_template:
        # never cache template render
        use_cache = False
    else:
        # archive is a input local file
        cache_targets = ['local']
        if not upstream:
            # project source is an input
            cache_targets += ['source']
        use_cache = proj.cache.enabled(
            *cache_targets, cmd='srcpkg', use_cache=cache)

    if not release:
        release = '1'

    if not in_format or in_format == 'auto':
        # default to list format when inputs arg is used, otherwise use yaml format
        in_format = 'list' if inputs else 'yaml'

    inputs = common.parse_inputs(inputs, in_files, in_format=in_format)

    if in_format == 'list':
        # convert inputs from list to yaml format
        if inputs:
            inputs = {'archive': inputs[0]}
        else:
            inputs = {}

    sanitize_archive_output(inputs)

    if not archive:
        # archive not specified - use make_archive or get_archive
        if inputs:
            raise ex.InvalidInput(
                fail="unexpected input:\n\n%s" % common.yaml_dump(inputs))

        if upstream:
            inputs = get_archive(
                version=version,
                cache=cache,
                project=proj)
        else:
            inputs = make_archive(
                cache=cache,
                project=proj)

    ar_path = inputs.get('archive')
    if not ar_path:
        raise ex.InvalidInput(
            msg="Missing required input: archive")
    ar_path = Path(ar_path)

    version = inputs.get('version')
    if not version:
        version = get_archive_version(ar_path)

    components = inputs.get('components', {})


    paths = [ar_path] + list(components.values())
    common.ensure_inputs(paths)

    if use_cache:
        cache_key = 'srcpkg/%s/%s/%s-%s/' % (
            srcpkg_type, distro.idver, version, release)
        cache_key += '%s:%s' % (
            path_checksum(*paths), file_checksum(*paths))
        if not upstream:
            # dev srcpkg uses project source as input
            cache_key += ':%s' % proj.checksum
        cached = common.get_cached_paths(proj, cache_key, result_dir)
        if cached:
            log.success("reuse cached source package: %s", cached[0])
            return cached

    if upstream:
        # --upstream mode - use distro/ from archive
        proj.load_upstream_archive(ar_path)

    # fetch correct package template
    template = proj.get_template_for_distro(distro)
    if not template:
        tdir = proj.path.templates
        msg = ("missing package template for distro: %s\n\n"
               "you can add it into: %s" % (distro, tdir))
        raise ex.MissingPackagingTemplate(msg=msg)
    ps = template.pkgstyle
    log.info("package style: %s", ps.name)
    log.info("package template: %s", template.path)
    log.info("package archive: %s", ar_path)

    # get needed paths
    try:
        pkg_name = call_pkgstyle_fun(
            template.pkgstyle, 'get_template_name',
            template, distro=distro)
    except ex.ApkgException as e:
        pkg_name = proj.name
        log.info("%s, using project.name instead: %s", str(e), pkg_name)
    nvr = "%s-%s-%s" % (pkg_name, version, release)
    nvr = common.sanitize_fn(nvr)
    build_path = proj.path.srcpkg_build / distro.idver / nvr
    if result_dir:
        out_path = Path(result_dir)
    else:
        out_path = proj.path.srcpkg_out / distro.idver / nvr
    log.info("package NVR: %s", nvr)
    log.info("build dir: %s", build_path)
    log.info("result dir: %s", out_path)

    # prepare new build dir
    if build_path.exists():
        log.info("removing existing build dir: %s", build_path)
        shutil.rmtree(build_path)
    build_path.mkdir(parents=True, exist_ok=True)
    # ensure output dir doesn't exist unless it was specified
    if not result_dir and out_path.exists():
        log.info("removing existing result dir: %s", out_path)
        shutil.rmtree(out_path)

    # prepare vars accessible from templates
    tvars = {
        'name': pkg_name,
        'version': version,
        'release': release,
        'nvr': nvr,
        'distro': distro,
    }
    if render_template:
        # render template only, don't build srcpkg
        if result_dir:
            # respect --result-dir when rendering template
            build_path = out_path
        template.render(build_path, tvars=tvars)
        log.success("rendered source package template: %s", build_path)
        return [build_path]

    # create source package using desired package style
    results = template.pkgstyle.build_srcpkg(
        build_path,
        out_path,
        archive_info=inputs,
        template=template,
        tvars=tvars)

    # check reported results exist
    for p in results:
        if not Path(p).exists():
            msg = ("source package build reported success but result is "
                   "missing:\n\n%s" % p)
            raise ex.UnexpectedCommandOutput(msg=msg)

    # first line of output is a source package by convention
    srcpkg_path, *rest = results
    # sort additional srcpkg files for determinism
    results = [srcpkg_path] + sorted(rest)

    log.success("made source package: %s", srcpkg_path)

    if use_cache:
        proj.cache.update(cache_key, results)

    return results


def sanitize_inputs(inputs):
    # convert paths to pathlib.Path
    archive = inputs.get('archive')
    if archive and not isinstance(archive, Path):
        inputs['archive'] = Path(archive)
    components = inputs.get('components')
    if components:
        for key, val in list(components.items()):
            if not isinstance(val, Path):
                components[key] = Path(val)


APKG_CLI_COMMANDS = [cli_srcpkg]
